import os
import os.path
import math
import json
from pc23 import xrange
from .base import AttributeValueVirtual, ValueArrayAbstract, AttributeValueNull
from ..exceptions import OrmValueError, OrmUnacceptableOperation, OrmNotDefinedYet, OrmParameterError
from ..manage import manager


class ValueList(ValueArrayAbstract, list):
    def __init__(self):
        super(ValueList, self).__init__()
        self._objects = list()
        self._original_values = list()
        self._values = list()

    def __repr__(self):
        return "List(%r)" % self._values

    def _update_objects_pos(self):
        sz = len(self._objects)
        if sz > 0:
            for n in xrange(sz):
                self._objects[n].pos = n

    def reset(self):
        self._objects = list()
        self._to_delete = list()
        self._loaded = False
        self._values = list()

    def write(self):
        self._update_objects_pos()
        super(ValueList, self).write()

    def load(self):
        from ..funcs import where_
        c = self._get_c()
        if c is None or self.instance is None:
            return
        self._loaded = True
        model = self.instance.__class__
        session = manager.get_session(instance=self.instance)
        pk = model.__meta__.primary_key
        w = list()
        for k in pk:
            w.append(getattr(c.array_model, 'parent_%s' % k) == getattr(self.instance, k))
        objects = session.select(c.array_model).where(where_(*w)).order_by(getattr(c.array_model, 'pos')).go()
        self._values = list()
        for o in objects:
            value = getattr(o, 'value')
            self._values.append(value)
            self._objects.append(o)
        self._update_objects_pos()
        self._original_values = list(self._values)

    def __contains__(self, x):
        self.ensure()
        return x in self._values

    def __delitem__(self, i):
        self.delete(i)

    def __delslice__(self, i, j):
        # TODO !
        pass

    def __getitem__(self, i):
        return self.get(i)

    def __getslice__(self, i, j):
        # TODO !
        pass

    def __iter__(self):
        self.ensure()
        for x in self._values:
            yield x

    def __len__(self):
        self.ensure()
        return 0 if self._values is None else len(self._values)

    def __reversed__(self):
        return self._values.__reversed__()

    def __setitem__(self, i, x):
        self.set(i, x)

    def append(self, x):
        self.ensure()
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot access values of List until model been defined!")
        if c.unique and x in self._values:
            return
        obj = self.c.array_model()
        self._assign_with_instance(obj)
        obj.value = x
        self._objects.append(obj)
        self._values.append(x)
        self._update_objects_pos()
        self._verify_changes()

    def delete(self, i):
        self.ensure()
        if i >= len(self._values):
            raise IndexError(i)
        obj = self._objects[i]
        self._to_delete.append(obj)
        del (self._values[i])
        del (self._objects[i])
        self._update_objects_pos()
        self._verify_changes()

    def remove(self, x):
        self.ensure()
        if x not in self._values:
            raise KeyError(x)
        self.delete(self._values.index(x))

    def set(self, i, x):
        self.ensure()
        if i >= len(self._values):
            raise IndexError(i)
        i_ = self._values.index(x)
        if i_ and i_ != i:
            raise OrmValueError("value is already exists in the unique List: '%r'!" % x)
        self._values[i] = x
        self._objects[i].value = x
        self._verify_changes()

    def get(self, i):
        self.ensure()
        if i >= len(self._values):
            raise IndexError(i)
        return self._values[i]

    def count(self, x):
        self.ensure()
        return self._values.count(x)

    def extend(self, iterable):
        for x in iterable:
            self.append(x)

    def index(self, x, start=None, stop=None):
        return self._values.index(x, start, stop)

    def insert(self, i, x):
        self.ensure()
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot access values of List until model been defined!")
        if c.unique and x in self._values:
            return
        obj = self.c.array_model()
        self._assign_with_instance(obj)
        obj.value = x
        self._objects.insert(i, obj)
        self._values.insert(i, x)
        self._update_objects_pos()
        self._verify_changes()

    def pop(self, i: int = -1):
        raise OrmUnacceptableOperation("cannot 'pop()' from the List!")

    def reverse(self):
        self.ensure()
        self._values.reverse()
        self._objects.reverse()
        self._update_objects_pos()
        self._verify_changes()

    def sort(self, cmp=None, key=None, reverse=False):
        raise OrmUnacceptableOperation("cannot 'sort()' List attribute!")


class ValueDict(ValueArrayAbstract, dict):
    def __init__(self):
        super(ValueDict, self).__init__()
        self._objects = dict()
        self._original_values = dict()
        self._values = dict()

    def __repr__(self):
        return "Dict(%r)" % self._values

    def load(self):
        from ..funcs import where_
        c = self._get_c()
        if c is None or self.instance is None:
            return
        self._loaded = True
        model = self.instance.__class__
        session = manager.get_session(instance=self.instance)
        pk = model.__meta__.primary_key
        w = list()
        for k in pk:
            w.append(getattr(c.array_model, 'parent_%s' % k) == getattr(self.instance, k))
        objects = session.select(c.array_model).where(where_(*w)).go()
        self._values = dict()
        for o in objects:
            key = getattr(o, '_k_')
            value = getattr(o, '_v_')
            self._values[key] = value
            self._objects[key] = o
        self._original_values = self._values.copy()

    def reset(self):
        self._objects = dict()
        self._to_delete = list()
        self._values = dict()
        self._loaded = False

    def get(self, key, default=None):
        self.ensure()
        return default if key not in self._values else self._values[key]

    def set(self, key, value):
        self.ensure()
        if key in self._values:
            obj = self._objects[key]
            setattr(obj, key, value)
        else:
            c = self._get_c()
            if c is None:
                raise OrmNotDefinedYet("cannot access values of Dict until model been defined!")
            obj = self.c.array_model()
            self._assign_with_instance(obj)
            obj._k_ = key
            obj._v_ = value
            self._objects[key] = obj
        self._values[key] = value
        self._verify_changes()

    def has(self, key):
        self.ensure()
        return key in self._values

    def delete(self, key):
        self.ensure()
        if key not in self._values:
            raise KeyError(key)
        obj = self._objects[key]
        self._to_delete.append(obj)
        del (self._objects[key])
        del (self._values[key])
        self._verify_changes()

    def __contains__(self, key):
        return self.has(key)

    def __delitem__(self, key):
        self.delete(key)

    def __getitem__(self, key):
        return self.get(key)

    def __iter__(self):
        self.ensure()
        for k in self._values:
            yield k

    def __len__(self):
        self.ensure()
        return 0 if self._values is None else len(self._values)

    def __setitem__(self, key, value):
        self.set(key, value)

    def clear(self):
        self.ensure()
        keys = self._values.keys()
        for k in keys:
            self.delete(k)

    def copy(self):
        self.ensure()
        return self._values.copy()

    def has_key(self, key):
        self.ensure()
        return self.has(key)

    def items(self):
        self.ensure()
        return self._values.items()

    def keys(self):
        self.ensure()
        return self._values.keys()

    def pop(self, key, default=None):
        raise OrmUnacceptableOperation("cannot 'pop()' from the Dict() attribute type!")

    def popitem(self):
        raise OrmUnacceptableOperation("cannot 'popitem()' from the Dict() attribute type!")

    def setdefault(self, key, default=None):
        self.ensure()
        if not self.has(key):
            self.set(key, default)
        return self.get(key)

    def update(self, **kwargs):
        self.ensure()
        for k in kwargs:
            v = kwargs[k]
            self.set(k, v)

    def values(self):
        self.ensure()
        return self._values.values()


class ValueMatrixRow(dict):
    """ The class representing a single row of values of the Matrix attribute. While it is basing on the
    dict, the difference is that this row disallows deletion or adding of keys (the set of keys is basing
    on the defined attribute's columns) and all changes in the row automatically are reflected in the
    corresponding real database object. """

    def __init__(self):
        super(ValueMatrixRow, self).__init__()
        self.instance = None
        self.model_k = None
        self.c = None
        self.obj = None

    def __setitem__(self, key, value):
        if key not in self.c.columns:
            raise KeyError(key)
        super(ValueMatrixRow, self).__setitem__(key, value)
        setattr(self.obj, key, value)

    def __contains__(self, key):
        return key in self.c.columns

    def __delitem__(self, key):
        raise OrmUnacceptableOperation("cannot delete key the of matrix row!")

    def clear(self):
        raise OrmUnacceptableOperation("cannot 'clear()' the of matrix row!")

    def has_key(self, key):
        return key in self.c.columns

    def pop(self, key, default=None):
        raise OrmUnacceptableOperation("cannot 'pop()' from the Dict() attribute type!")

    def popitem(self):
        raise OrmUnacceptableOperation("cannot 'popitem()' from the Dict() attribute type!")

    def setdefault(self, key, default=None):
        if key not in self.c.columns:
            raise KeyError(key)
        return super(ValueMatrixRow, self).__getitem__(key)

    def update(self, **kwargs):
        for k in kwargs:
            self.__setitem__(k, kwargs[k])


class ValueMatrix(ValueDict):
    def __init__(self):
        super(ValueMatrix, self).__init__()
        self._objects = dict()
        self._original_values = dict()
        self._values = dict()

    def __repr__(self):
        return "Matrix(%r)" % self._values

    def _matrixrow_from_object(self, obj):
        c = self._get_c()
        row = ValueMatrixRow()
        row.obj = obj
        row.instance = self.instance
        row.model_k = self.model_k
        row.c = c
        for k in c.columns:
            row[k] = getattr(obj, k)
        return row

    def load(self):
        from ..funcs import where_
        c = self._get_c()
        if c is None or self.instance is None:
            return
        self._loaded = True
        model = self.instance.__class__
        session = manager.get_session(instance=self.instance)
        pk = model.__meta__.primary_key
        w = list()
        for k in pk:
            w.append(getattr(c.array_model, 'parent_%s' % k) == getattr(self.instance, k))
        objects = session.select(c.array_model).where(where_(*w)).go()
        self._values = dict()
        for o in objects:
            row = self._matrixrow_from_object(o)
            key = getattr(o, '_k_')
            self._values[key] = row
            self._objects[key] = o
        self._original_values = self._values.copy()

    def reset(self):
        self._objects = dict()
        self._to_delete = list()
        self._values = dict()
        self._loaded = False

    def get(self, key, default=None):
        self.ensure()
        return default if key not in self._values else self._values[key]

    def set(self, key, value):
        if not isinstance(value, dict):
            raise OrmValueError("Matrix attribute requires (dict) type to be given when setting a whole row!")
        self.ensure()
        if key in self._values:
            obj = self._objects[key]
            for k in value:
                if k not in self._get_c().column:
                    raise KeyError("matrix attribute has no column '%s'!" % k)
                setattr(obj, k, value[k])
        else:
            c = self._get_c()
            if c is None:
                raise OrmNotDefinedYet("cannot access values of Matrix until model been defined!")
            obj = self.c.array_model()
            self._assign_with_instance(obj)
            obj._k_ = key
            for k in value:
                if k not in self._get_c().columns:
                    raise KeyError("matrix attribute has no column '%s'!" % k)
                setattr(obj, k, value[k])
            self._objects[key] = obj
        self._values[key] = self._matrixrow_from_object(obj)
        self._verify_changes()

    def has(self, key):
        self.ensure()
        return key in self._values

    def delete(self, key):
        self.ensure()
        if key not in self._values:
            raise KeyError(key)
        obj = self._objects[key]
        self._to_delete.append(obj)
        del (self._objects[key])
        del (self._values[key])
        self._verify_changes()

    def __contains__(self, key):
        return self.has(key)

    def __delitem__(self, key):
        self.delete(key)

    def __getitem__(self, key):
        return self.get(key)

    def __iter__(self):
        self.ensure()
        for k in self._values:
            yield k

    def __len__(self):
        self.ensure()
        return 0 if self._values is None else len(self._values)

    def __setitem__(self, key, value):
        self.set(key, value)

    def clear(self):
        self.ensure()
        keys = self._values.keys()
        for k in keys:
            self.delete(k)

    def copy(self):
        self.ensure()
        return self._values.copy()

    def has_key(self, key):
        self.ensure()
        return self.has(key)

    def items(self):
        self.ensure()
        return self._values.items()

    def keys(self):
        self.ensure()
        return self._values.keys()

    def pop(self, key, default=None):
        raise OrmUnacceptableOperation("cannot 'pop()' from the Matrix() attribute type!")

    def popitem(self):
        raise OrmUnacceptableOperation("cannot 'popitem()' from the Matrix() attribute type!")

    def setdefault(self, key, default=None):
        self.ensure()
        if not self.has(key):
            self.set(key, default)
        return self.get(key)

    def update(self, **kwargs):
        self.ensure()
        for k in kwargs:
            v = kwargs[k]
            self.set(k, v)

    def values(self):
        self.ensure()
        return self._values.values()


class ValueSet(ValueArrayAbstract, set):
    def __init__(self):
        super(ValueSet, self).__init__()
        self._objects = list()
        self._original_values = set()
        self._values = set()

    def __repr__(self):
        return "Set(%r)" % sorted(list(self._values))

    def reset(self):
        self._objects = list()
        self._loaded = False
        self._values = set()

    def write(self):
        dbobjs = self._load_from_db()
        found = list()
        deleted = list()
        for o in dbobjs:
            opt = o.value
            if opt in found or opt not in self._values:
                o.delete()
                deleted.append(o)
                continue
            found.append(opt)
        for o in self._objects:
            opt = o.value
            if opt not in self._values or opt in found:
                continue
            o.write()

    def _load_from_db(self):
        from ..funcs import where_
        c = self._get_c()
        if c is None or self.instance is None:
            return
        model = self.instance.__class__
        session = manager.get_session(instance=self.instance)
        pk = model.__meta__.primary_key
        w = list()
        for k in pk:
            w.append(getattr(c.array_model, 'parent_%s' % k) == getattr(self.instance, k))
        return session.select(c.array_model).where(where_(*w)).go()

    def load(self):
        c = self._get_c()
        if c is None or self.instance is None:
            return
        self._loaded = True
        objects = self._load_from_db()
        self._values = set()
        for o in objects:
            value = getattr(o, 'value')
            self._values.add(value)
            self._objects.append(o)
        self._original_values = set(self._values)

    def __contains__(self, x):
        self.ensure()
        x = str(x)
        return x in self._values

    def __iter__(self):
        self.ensure()
        for x in self._values:
            yield x

    def __len__(self):
        self.ensure()
        return 0 if self._values is None else len(self._values)

    def _ensure_obj(self, x):
        self.ensure()
        x = str(x)
        if x not in self.c.options:
            raise OrmParameterError("value '%s' is not in options list for this Set!" % x)
        for o in self._objects:
            if o.value == x:
                return
        obj = self.c.array_model()
        self._assign_with_instance(obj)
        obj.value = x
        self._objects.append(obj)

    def add(self, x):
        self.ensure()
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot access values of Set until model been defined!")
        x = str(x)
        if x in self._values:
            return
        self._ensure_obj(x)
        self._values.add(x)
        self._verify_changes()

    def remove(self, x):
        self.ensure()
        x = str(x)
        if x not in self._values:
            raise KeyError(x)
        self._values.remove(x)
        self._verify_changes()

    def discard(self, x):
        self.ensure()
        x = str(x)
        if x not in self._values:
            return
        self._values.remove(x)
        self._verify_changes()

    def clear(self):
        self.ensure()
        self._values = set()
        self._verify_changes()

    def copy(self):
        self.ensure()
        return self._values.copy()

    def difference(self, s):
        self.ensure()
        return self._values.difference(set(str(x) for x in s))

    def difference_update(self, s):
        self.ensure()
        for x in s:
            if str(x) not in self._values:
                continue
            self.remove(x)
        return self

    def intersection(self, s):
        self.ensure()
        return self._values.intersection(set(str(x) for x in s))

    def intersection_update(self, s):
        self.ensure()
        for x in self._values:
            if str(x) not in s:
                continue
            self.remove(x)
        return self

    def isdisjoint(self, s):
        self.ensure()
        return self._values.isdisjoint(set(str(x) for x in s))

    def issubset(self, s):
        self.ensure()
        return self._values.issubset(set(str(x) for x in s))

    def issuperset(self, s):
        self.ensure()
        return self._values.issuperset(set(str(x) for x in s))

    def pop(self):
        raise OrmUnacceptableOperation("cannot 'pop()' from the Set() attribute type!")

    def symmetric_difference(self, s):
        self.ensure()
        return self._values.symmetric_difference(set(str(x) for x in s))

    def symmetric_difference_update(self, s):
        self.ensure()
        s_ = self._values.copy()
        for x in self._values:
            if not(str(x) in s and str(x) in s_):
                continue
            self.remove(x)
        for x in s:
            if str(x) in s and str(x) in s_:
                continue
            self.add(str(x))
        return self

    def union(self, s):
        self.ensure()
        for x in s:
            if str(x) in self._values:
                continue
            self.add(str(x))

    def update(self, s):
        self.ensure()
        for x in s:
            self.add(x)


class ValueInlineSet(AttributeValueVirtual, set):
    basecls = False
    check_modified = True

    def __init__(self, value):
        super(ValueInlineSet, self).__init__()
        if value is None:
            value = set()
        elif isinstance(value, str):
            value = self._from_text(value)
        self._value = value
        self._original_value = self._value

    def _get_c(self):
        return getattr(self.instance.__class__, self.model_k) if self.model_k is not None else None

    def _verify_changes(self):
        if self._original_value == self._value and self.model_k in self.instance.__changes__:
            del(self.instance.__changes__[self.model_k])
        elif self._original_value != self._value:
            self.instance.__changes__[self.model_k] = (self._original_value, self._value)

    def __repr__(self):
        return "InlineSet(%r)" % sorted(list(self._value))

    @staticmethod
    def _to_text(value):
        return json.dumps(list(value))

    @staticmethod
    def _from_text(value):
        return set(json.loads(value))

    def base(self, value):
        return self._value

    @property
    def value(self):
        return self._value

    @property
    def dbvalue(self):
        return self._to_text(self._value)

    @property
    def string(self):
        return self.__repr__()

    def get_value(self):
        return self

    def set_value(self, instance, key, value):
        before_ = self._value
        if isinstance(value, ValueInlineSet):
            self._value = value.value
            return self._value, before_
        if value is None or isinstance(value, AttributeValueNull):
            self._value = set()
            return self._value, before_
        if isinstance(value, (set, frozenset, list, tuple)):
            self._value = set(value)
            return self._value, before_
        self._value = set()
        self._value.add(str(value))
        return self._value, before_

    def __contains__(self, x):
        x = str(x)
        return x in self._value

    def __iter__(self):
        for x in self._value:
            yield x

    def __len__(self):
        return 0 if self._value is None else len(self._value)

    def add(self, x):
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot access values of InlineSet until model been defined!")
        x = str(x)
        if x not in self.c.options:
            raise OrmParameterError("value '%s' is not in options list for this Set!" % x)
        if x in self._value:
            return
        self._value.add(x)
        self._verify_changes()

    def remove(self, x):
        x = str(x)
        if x not in self._value:
            raise KeyError(x)
        self._value.remove(x)
        self._verify_changes()

    def discard(self, x):
        x = str(x)
        if x not in self._value:
            return
        self._value.remove(x)

    def clear(self):
        self._value = set()
        self._verify_changes()

    def copy(self):
        return self._value.copy()

    def difference(self, s):
        return self._value.difference(set(str(x) for x in s))

    def difference_update(self, s):
        for x in s:
            if str(x) not in self._value:
                continue
            self.remove(x)
        self._verify_changes()
        return self

    def intersection(self, s):
        return self._value.intersection(set(str(x) for x in s))

    def intersection_update(self, s):
        for x in self._value:
            if str(x) not in s:
                continue
            self.remove(x)
        self._verify_changes()
        return self

    def isdisjoint(self, s):
        return self._value.isdisjoint(set(str(x) for x in s))

    def issubset(self, s):
        return self._value.issubset(set(str(x) for x in s))

    def issuperset(self, s):
        return self._value.issuperset(set(str(x) for x in s))

    def pop(self):
        return self._value.pop()

    def symmetric_difference(self, s):
        return self._value.symmetric_difference(set(str(x) for x in s))

    def symmetric_difference_update(self, s):
        s_ = self._value.copy()
        for x in self._value:
            if not(str(x) in s and str(x) in s_):
                continue
            self.remove(x)
        for x in s:
            if str(x) in s and str(x) in s_:
                continue
            self.add(str(x))
        self._verify_changes()
        return self

    def union(self, s):
        for x in s:
            if str(x) in self._value:
                continue
            self.add(str(x))
        self._verify_changes()

    def update(self, s):
        for x in s:
            self.add(x)
        self._verify_changes()


class ValueAttachmentsAbstract(ValueArrayAbstract, list):
    def __init__(self):
        super(ValueAttachmentsAbstract, self).__init__()
        self._objects = list()
        self._original_values = list()
        self._values = list()

    def __repr__(self):
        return "Attachments(%r)" % self._values

    def reset(self):
        self._objects = list()
        self._to_delete = list()
        self._loaded = False
        self._values = list()

    def write(self):
        super(ValueAttachmentsAbstract, self).write()

    def _assign_with_instance(self, obj):
        super(ValueAttachmentsAbstract, self)._assign_with_instance(obj)
        obj.attachment.parent_instance = self.instance
        obj.attachment.parent_k = self.model_k

    def load(self):
        from ..funcs import where_
        c = self._get_c()
        if c is None or self.instance is None:
            return
        self._loaded = True
        model = self.instance.__class__
        session = manager.get_session(instance=self.instance)
        pk = model.__meta__.primary_key
        w = list()
        for k in pk:
            w.append(getattr(c.array_model, 'parent_%s' % k) == getattr(self.instance, k))
        objects = session.select(c.array_model).where(where_(*w)).order_by(c.array_model.time_created).go()
        self._values = list()
        for o in objects:
            o.attachment.parent_instance = self.instance
            o.attachment.parent_k = self.model_k
            o.attachment.attribute_instance = o
            o.attachment.verify_storage_existance()
            if not o.attachment.is_attached:
                self._to_delete.append(o)
                continue
            self._values.append(o.attachment.filename)
            self._objects.append(o)
        self._original_values = list(self._values)

    def after_parent_instance_deleted(self):
        self.clear()
        self.write()

    def __contains__(self, x):
        self.ensure()
        return x in self._values

    def __delitem__(self, i):
        self.delete_file(i)

    def __delslice__(self, i, j):
        # TODO !
        pass

    def __getitem__(self, i):
        self.ensure()
        if i >= len(self._values):
            raise IndexError(i)
        return self._objects[i]

    def __getslice__(self, i, j):
        # TODO !
        pass

    def __iter__(self):
        self.ensure()
        for x in self._objects:
            yield x

    def __len__(self):
        self.ensure()
        return 0 if self._values is None else len(self._values)

    def __reversed__(self):
        raise OrmUnacceptableOperation("cannot reverse the Files|Pictures attribute!")

    def __setitem__(self, i, x):
        raise OrmUnacceptableOperation("cannot set the Files|Pictures attribute by index to the some value!")

    def append(self, x):
        raise OrmUnacceptableOperation("cannot 'append()' to the Files|Pictures attribute!")

    def add_file_from(self, src, filename=None, rule=None, **kwargs):
        self.ensure()
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot access values of Files|Pictures until model been defined!")
        obj = self.c.array_model()
        self._assign_with_instance(obj)
        obj.attachment.push_file(src, filename, rule)
        obj.title = kwargs.pop('title', filename or obj.attachment.filename)
        obj.filesize = int(math.ceil(float(os.stat(obj.attachment.abs_path).st_size) / 1024.0))
        for k in kwargs:
            setattr(obj, k, kwargs[k])
        self._objects.append(obj)
        self._values.append(obj.attachment.filename)
        self._verify_changes()
        return obj

    def get_file_to(self, i, dst, rule=None):
        self.ensure()
        if isinstance(i, int) and i >= len(self._values):
            raise IndexError(i)
        elif isinstance(i, str):
            if i not in self._values:
                raise KeyError(i)
            i = self._values.index(i)
        elif not isinstance(i, (int, str)):
            raise OrmParameterError("must be (int|str) type!")
        self._objects[i].attachment.pull_file(dst, rule)

    def delete_file(self, i):
        self.ensure()
        if isinstance(i, int) and i >= len(self._values):
            raise IndexError(i)
        elif isinstance(i, str):
            if i not in self._values:
                raise KeyError(i)
            i = self._values.index(i)
        elif not isinstance(i, (int, str)):
            raise OrmParameterError("must be (int|str) type!")
        obj = self._objects[i]
        self._to_delete.append(obj)
        del (self._values[i])
        del (self._objects[i])
        self._verify_changes()

    def clear(self):
        self.ensure()
        while self.__len__() > 0:
            self.delete_file(0)

    def count(self, x):
        self.ensure()
        return self._values.count(x)

    def extend(self, iterable):
        raise OrmUnacceptableOperation("cannot 'extend()' the Files|Pictures attribute!")

    def index(self, x, start=None, stop=None):
        return self._values.index(x, start, stop)

    def insert(self, i, x):
        raise OrmUnacceptableOperation("cannot 'insert()' the Files|Pictures attribute!")

    def pop(self, i: int = -1):
        raise OrmUnacceptableOperation("cannot 'pop()' from the Files|Pictures!")

    def reverse(self):
        raise OrmUnacceptableOperation("cannot 'reverse()' the Files|Pictures!")

    def sort(self, cmp=None, key=None, reverse=False):
        raise OrmUnacceptableOperation("cannot 'sort()' the Files|Pictures attribute!")


class ValueFiles(ValueAttachmentsAbstract):
    def __init__(self):
        super(ValueFiles, self).__init__()

    def __repr__(self):
        return "Files(%r)" % self._values


class ValuePictures(ValueAttachmentsAbstract):
    def __init__(self):
        super(ValuePictures, self).__init__()

    def __repr__(self):
        return "Pictures(%r)" % self._values

    def add_image(self, image, filename=None, **kwargs):
        self.ensure()
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot access values of Files|Pictures until model been defined!")
        obj = self.c.array_model()
        self._assign_with_instance(obj)
        obj.attachment.push_image(image, filename, **kwargs)
        obj.title = kwargs.pop('title', filename or obj.attachment.filename)
        obj.filesize = None
        for k in kwargs:
            setattr(obj, k, kwargs[k])
        self._objects.append(obj)
        self._values.append(obj.attachment.filename)
        self._verify_changes()
        return obj

    def add_image_from(self, src, filename=None, **kwargs):
        self.ensure()
        c = self._get_c()
        if c is None:
            raise OrmNotDefinedYet("cannot access values of Files|Pictures until model been defined!")
        obj = self.c.array_model()
        self._assign_with_instance(obj)
        obj.attachment.push_image_from(src, filename, **kwargs)
        obj.title = kwargs.pop('title', filename or obj.attachment.filename)
        obj.filesize = int(math.ceil(float(os.stat(obj.attachment.abs_path).st_size) / 1024.0))
        for k in kwargs:
            setattr(obj, k, kwargs[k])
        self._objects.append(obj)
        self._values.append(obj.attachment.filename)
        self._verify_changes()
        return obj

    def get_image_to(self, i, dst):
        self.ensure()
        if isinstance(i, int) and i >= len(self._values):
            raise IndexError(i)
        elif isinstance(i, str):
            if i not in self._values:
                raise KeyError(i)
            i = self._values.index(i)
        elif not isinstance(i, (int, str)):
            raise OrmParameterError("must be (int|str) type!")
        self._objects[i].attachment.pull_image_to(dst)

    def get_image(self, i):
        self.ensure()
        if isinstance(i, int) and i >= len(self._values):
            raise IndexError(i)
        elif isinstance(i, str):
            if i not in self._values:
                raise KeyError(i)
            i = self._values.index(i)
        elif not isinstance(i, (int, str)):
            raise OrmParameterError("must be (int|str) type!")
        self._objects[i].attachment.pull_image()

    def delete_image(self, i):
        self.delete_file(i)



import datetime
import calendar
from pc23 import strlen
from .base import ColumnAttributeAbstract, OrmComparator
from ..values import AttributeValueInteger, AttributeValueFloat, AttributeValueString, AttributeValueBoolean, \
    AttributeValueBinary, AttributeValueDate, AttributeValueTime, AttributeValueDateTime, AttributeValueNull
from ..manage import manager
from ..utils import format_datetime, gettext
from ..exceptions import OrmValueVerificationError, OrmDefinitionError, OrmComparatorValueError


class ColumnIntegerAbstract(ColumnAttributeAbstract):
    def __init__(self, length, **kwargs):
        self.length = length
        self.big = kwargs.pop('big', False)
        self.unsigned = kwargs.pop('unsigned', False)
        self.require_nonzero = kwargs.pop('require_nonzero', False)
        if self.require_nonzero:
            kwargs['required'] = True
        super(ColumnIntegerAbstract, self).__init__(**kwargs)
        self.data_type = 'int'
        self.value_class = AttributeValueInteger
        if not self.nullable and not self._default and not self.auto_increment:
            self.default = 0
        self.override_null = 0
        self.strict_find = self.strict_find if self.strict_find is not None else True
        if self.is_version_key:
            self.nullable = bool(self.is_softdelete_key)
            self._default = 1
            self.unsigned = True

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if (not self.big and str(struct['data_type']).lower() != 'int') \
                    or (self.big and str(struct['data_type']).lower() != 'bigint'):
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and int(self._default) != int(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
            if self.auto_increment and 'auto_increment' not in str(struct['extra']).lower():
                return False
            if self.unsigned and 'unsigned' not in str(struct['column_type']).lower():
                return False
        elif session.dbms_type == 'postgresql':
            if (not self.big and str(struct['data_type']).lower() != 'integer') \
                    or (self.big and str(struct['data_type']).lower() != 'bigint'):
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if self.auto_increment:
                nextval = "nextval('%s_%s_seq'::regclass)" % (self.model.table, self.tbl_k)
                if struct['column_default'] != nextval:
                    return False
            elif not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and int(self._default) != int(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        if session.dbms_type == 'postgresql':
            data_type = ("BIGSERIAL" if self.big else "SERIAL") \
                if self.auto_increment \
                else ("BIGINT" if self.big else "INT")
        else:
            data_type = "BIGINT" if self.big else "INT"
        d.append(data_type)
        if self.unsigned and session.dbms_type == 'mysql':
            d.append('UNSIGNED')
        if not self.nullable and not self.auto_increment:
            d.append('NOT NULL')
        if self.auto_increment:
            if session.dbms_type == 'mysql':
                d.append('AUTO_INCREMENT')
        elif not callable(self._default) and self._default is not None:
            d.append('DEFAULT %i' % int(self._default))
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            data_type = ("BIGSERIAL" if self.big else "SERIAL") \
                if self.auto_increment \
                else ("BIGINT" if self.big else "INT")
            d.append("SET DATA TYPE %s" % data_type)
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            if not callable(self.default) and self.default is not None and not self.auto_increment:
                d.append("SET DEFAULT %i" % self._default)
            else:
                d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def verify(self, value):
        value = super(ColumnIntegerAbstract, self).verify(value)
        if value is None:
            return None
        if '-' in str(value) and (str(value).index('-') != 0 or str(value).count('-') > 1):
            raise OrmValueVerificationError(gettext("Must be a number"))
        if not str(value).replace('-', '').isdigit():
            raise OrmValueVerificationError(gettext("Must be a number"))
        if self.unsigned and int(value) < 0:
            raise OrmValueVerificationError(gettext("Must be a positive number"))
        if self.require_nonzero and value == 0:
            raise OrmValueVerificationError(gettext("Is required"))
        return int(value)

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', '0'))

    @property
    def notempty_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', '0'))

    @property
    def true_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', '0'))

    @property
    def false_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', '0'))


class ColumnFloatAbstract(ColumnAttributeAbstract):
    def __init__(self, length, decimals, **kwargs):
        if decimals is None and length is not None:
            raise OrmDefinitionError(
                "floating-point attribute types cannot be defined with 'length', but without 'decimal'!")
        elif decimals is not None and length is None:
            raise OrmDefinitionError(
                "floating-point attribute types cannot be defined with 'decimal', but without 'length'!")
        self.length = length
        self.decimals = decimals
        self.unsigned = bool(kwargs.pop('unsigned', False))
        self.big = kwargs.pop('big', False)
        self.require_nonzero = bool(kwargs.pop('require_nonzero', False))
        if self.require_nonzero:
            kwargs['required'] = True
        super(ColumnFloatAbstract, self).__init__(**kwargs)
        self.data_type = 'float'
        self.value_class = AttributeValueFloat
        if not self.nullable and not self._default:
            self._default = 0.0
        self.override_null = 0.0
        self.strict_find = self.strict_find if self.strict_find is not None else True

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if (not self.big and str(struct['data_type']).lower() != 'float') \
                    or (self.big and str(struct['data_type']).lower() != 'double'):
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and float(self._default) != float(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
            if self.unsigned and 'unsigned' not in str(struct['column_type']).lower():
                return False
        elif session.dbms_type == 'postgresql':
            if (not self.big and str(struct['data_type']).lower() != 'real') \
                    or (self.big and str(struct['data_type']).lower() != 'double precision'):
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and float(self._default) != float(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        if session.dbms_type == 'postgresql':
            data_type = "DOUBLE PRECISION" if self.big else "REAL"
        elif session.dbms_type == 'mysql':
            data_type = "DOUBLE" if self.big else "FLOAT"
            if self.length and self.decimals:
                data_type += "(%i,%i)" % (int(self.length), int(self.decimals))
        else:
            raise TypeError("this column type is not supported for this database and has no alternative!")
        d.append(data_type)
        if self.unsigned and session.dbms_type == 'mysql':
            d.append('UNSIGNED')
        if not self.nullable:
            d.append('NOT NULL')
        if not callable(self._default) and self._default is not None:
            d.append('DEFAULT %f' % float(self._default))
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            data_type = "DOUBLE PRECISION" if self.big else "REAL"
            d.append("SET DATA TYPE %s" % data_type)
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            if not callable(self.default) and self.default is not None:
                d.append("SET DEFAULT %f" % float(self._default))
            else:
                d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def verify(self, value):
        value = super(ColumnFloatAbstract, self).verify(value)
        if value is None:
            return None
        if '-' in str(value) and (str(value).index('-') != 0 or str(value).count('-') > 1):
            raise OrmValueVerificationError(gettext("Must be a floating-point number"))
        if str(value).count('.') > 1:
            raise OrmValueVerificationError(gettext("Must be a floating-point number"))
        if not str(value).replace('-', '').replace('.', '').isdigit():
            raise OrmValueVerificationError(gettext("Must be a floating-point number"))
        if self.unsigned and float(value) < 0:
            raise OrmValueVerificationError(gettext("Must be a positive floating-point number"))
        if self.require_nonzero and value == 0:
            raise OrmValueVerificationError(gettext("Is required"))
        return float(value)

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', '0'))

    @property
    def notempty_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', '0'))

    @property
    def true_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', '0'))

    @property
    def false_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', '0'))


class ColumnNumericAbstract(ColumnAttributeAbstract):
    def __init__(self, precision, scale, **kwargs):
        if not precision or not scale:
            raise OrmDefinitionError("numeric attribute types cannot be defined without 'precision' or 'scale'!")
        self.precision = int(precision)
        self.scale = int(scale)
        self.unsigned = bool(kwargs.pop('unsigned', False))
        self.require_nonzero = bool(kwargs.pop('require_nonzero', False))
        if self.require_nonzero:
            kwargs['required'] = True
        super(ColumnNumericAbstract, self).__init__(**kwargs)
        self.data_type = 'float'
        self.value_class = AttributeValueFloat
        if not self.nullable and not self._default:
            self._default = 0.0
        self.override_null = 0.0
        self.strict_find = self.strict_find if self.strict_find is not None else True

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() not in ('numeric', 'decimal'):
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and float(self._default) != float(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
            if self.unsigned and 'unsigned' not in str(struct['column_type']).lower():
                return False
            if int(self.precision) != int(struct['numeric_precision']):
                return False
            if int(self.scale) != int(struct['numeric_scale']):
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['data_type']).lower() != 'numeric':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and float(self._default) != float(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
            if int(self.precision) != int(struct['numeric_precision']):
                return False
            if int(self.scale) != int(struct['numeric_scale']):
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        if session.dbms_type == 'postgresql':
            data_type = "NUMERIC(%i,%i)" % (self.precision, self.scale)
        elif session.dbms_type == 'mysql':
            data_type = "DECIMAL(%i,%i)" % (self.precision, self.scale)
        else:
            raise TypeError("this column type is not supported for this database and has no alternative!")
        d.append(data_type)
        if self.unsigned and session.dbms_type == 'mysql':
            d.append('UNSIGNED')
        if not self.nullable:
            d.append('NOT NULL')
        if not callable(self._default) and self._default is not None:
            d.append('DEFAULT %f' % float(self._default))
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            data_type = "NUMERIC(%i,%i)" % (self.precision, self.scale)
            d.append("SET DATA TYPE %s" % data_type)
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            if not callable(self.default) and self.default is not None:
                d.append("SET DEFAULT %f" % float(self._default))
            else:
                d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', '0'))

    @property
    def notempty_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', '0'))

    @property
    def true_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', '0'))

    @property
    def false_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', '0'))


class ColumnDatetimeAbstract(ColumnAttributeAbstract):
    def __init__(self, fsp=None, **kwargs):
        self.fsp = fsp
        self.with_time_zone = bool(kwargs.pop('with_time_zone', True))
        super(ColumnDatetimeAbstract, self).__init__(**kwargs)
        self.data_type = 'datetime'
        self.value_class = AttributeValueDateTime
        self.strict_find = self.strict_find if self.strict_find is not None else True
        self.full_format = bool(kwargs.pop('full_format', False))
        self.format = kwargs.pop('format', None)
        self.use24hours = bool(kwargs.pop('use24hours', None))

    def represent_for_write(self, value, session):
        if self.with_time_zone and session.dbms_type == 'mysql' \
                and (value is None or isinstance(value, AttributeValueNull)):
            return '0000-00-00 00:00:00'
        return value

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() != ('timestamp' if self.with_time_zone else 'datetime'):
                return False
            if not self.with_time_zone and bool(struct['is_nullable']) != bool(self.nullable):
                return False
            if int(struct['datetime_precision']) != (int(self.fsp) if self.fsp is not None else 0):
                return False
            if self.with_time_zone \
                    and (struct['column_default'][:19] != '0000-00-00 00:00:00' or struct['extra']):
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['data_type']).lower() != ('timestamp with time zone'
                                                    if self.with_time_zone
                                                    else 'timestamp without time zone'):
                return False
            if bool(self.nullable) != bool(struct['is_nullable']):
                return False
            if int(struct['datetime_precision']) != (int(self.fsp) if self.fsp is not None else 6):
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        if session.dbms_type == 'postgresql':
            if self.with_time_zone:
                data_type = "TIMESTAMP WITH TIME ZONE" \
                    if not self.fsp \
                    else "TIMESTAMP(%i) WITH TIME ZONE" % int(self.fsp)
            else:
                data_type = "TIMESTAMP WITHOUT TIME ZONE" \
                    if not self.fsp \
                    else "TIMESTAMP(%i) WITHOUT TIME ZONE" % int(self.fsp)
        elif session.dbms_type == 'mysql':
            if self.with_time_zone:
                data_type = "TIMESTAMP" if not self.fsp else "TIMESTAMP(%i)" % int(self.fsp)
            else:
                data_type = "DATETIME" if not self.fsp else "DATETIME(%i)" % int(self.fsp)
        else:
            raise TypeError("this column type is not supported for this database and has no alternative!")
        d.append(data_type)
        if not self.nullable and (session.dbms_type != 'mysql' or not self.with_time_zone):
            d.append('NOT NULL')
        if session.dbms_type == 'mysql' and self.with_time_zone:
            d.append("DEFAULT 0")
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            if self.with_time_zone:
                data_type = "TIMESTAMP WITH TIME ZONE" \
                    if not self.fsp \
                    else "TIMESTAMP(%i) WITH TIME ZONE" % int(self.fsp)
            else:
                data_type = "TIMESTAMP WITHOUT TIME ZONE" \
                    if not self.fsp \
                    else "TIMESTAMP(%i) WITHOUT TIME ZONE" % int(self.fsp)
            d.append("SET DATA TYPE %s" % data_type)
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def text_repr(self, value, instance=None):
        if value is None:
            return ""
        session = manager.get_session(instance=instance)
        if isinstance(value, datetime.date) and not isinstance(value, datetime.datetime):
            format_ = self.format or \
                      (manager.get_full_date_format(session=session)
                       if self.full_format
                       else manager.get_short_date_format(session=session))
        elif isinstance(value, datetime.time):
            format_ = self.format or \
                      (manager.get_full_time_format(session=session)
                       if self.full_format
                       else manager.get_short_time_format(session=session))
        else:
            dateformat_ = self.format or \
                          (manager.get_full_date_format(session=session)
                           if self.full_format
                           else manager.get_short_date_format(session=session))
            timeformat_ = self.format or \
                          (manager.get_full_time_format(session=session)
                           if self.full_format
                           else manager.get_short_time_format(session=session))
            format_ = dateformat_ + " " + timeformat_
        return format_datetime(value, format_)

    @staticmethod
    def _verify_fail():
        raise OrmValueVerificationError(gettext("Must be a valid date and time"))

    def verify(self, value):
        value = super(ColumnDatetimeAbstract, self).verify(value)
        if value is None:
            return None
        if isinstance(value, str):
            ps = value.split(' ')
            d_ = False
            t_ = False
            n_year = 0
            n_month = 0
            n_day = 0
            n_hour = 0
            n_minute = 0
            n_second = 0
            for p in ps:
                if not p:
                    continue
                if '.' not in p and '/' not in p and '\\' not in p and '-' not in p and ':' not in p:
                    self._verify_fail()
                if ':' in p:
                    if t_:
                        self._verify_fail()
                    if p.count(':') > 2:
                        self._verify_fail()
                    p_ = p.split(':')
                    if not p_[0].isdigit() or not p_[1].isdigit() or (len(p_) == 3 and not p_[2].isdigit()):
                        self._verify_fail()
                    n_hour = int(p_[0])
                    n_minute = int(p_[1])
                    n_second = int(p_[2]) if len(p) == 3 else 0
                    if n_hour > 23 or n_minute > 59 or n_second > 59:
                        self._verify_fail()
                    t_ = True
                else:
                    if d_:
                        self._verify_fail()
                    s = '.' if '.' in p else ('-' if '-' in p else ('/' if '/' in p else '\\'))
                    p_ = p.split(s)
                    if len(p_) != 3:
                        self._verify_fail()
                    if not p_[0].isdigit() or not p_[1].isdigit() or not p_[2].isdigit():
                        self._verify_fail()
                    p1 = int(p_[0])
                    p2 = int(p_[1])
                    p3 = int(p_[2])
                    starting_with_year = p1 > 31 or s == '-'
                    if starting_with_year:
                        n_year = p1
                        if s in ('.', '-'):
                            n_month = p2
                            n_day = p3
                        else:
                            n_month = p3
                            n_day = p2
                    else:
                        n_year = p3
                        if s in ('.', '-'):
                            n_month = p2
                            n_day = p1
                        else:
                            n_month = p1
                            n_day = p2
                    if 99 < n_year < 1000:
                        self._verify_fail()
                    if n_year < 1000:
                        n_year = n_year + 1000 if n_year > 69 else n_year + 2000
                    if n_day == 0 or n_month == 0 or n_day > 31 or n_month > 12:
                        self._verify_fail()
                    n_daysmax = calendar.monthrange(n_year, n_month)[1]
                    if n_day > n_daysmax:
                        self._verify_fail()
                    d_ = True
            if not d_ or not t_:
                self._verify_fail()
            value = datetime.datetime(n_year, n_month, n_day, n_hour, n_minute, n_second)
        elif not isinstance(value, datetime.datetime):
            self._verify_fail()
        return value

    @property
    def empty_(self):
        return OrmComparator(self, 'FALSE')

    @property
    def notempty_(self):
        return OrmComparator(self, 'TRUE')

    @property
    def true_(self):
        return OrmComparator(self, 'TRUE')

    @property
    def false_(self):
        return OrmComparator(self, 'FALSE')

    def resolve_comparator_(self, cmp, session):
        from ..funcs import or_, and_
        if cmp.cmp == 'TRUE':
            if session.dbms_type == 'mysql':
                return and_(OrmComparator(cmp.item, '!=', '0000-00-00 00:00:00'), OrmComparator(cmp.item, '!ISNULL'))
            elif session.dbms_type == 'postgresql':
                return OrmComparator(cmp.item, '!ISNULL')
        elif cmp.cmp == 'FALSE':
            if session.dbms_type == 'mysql':
                return or_(OrmComparator(cmp.item, '=', '0000-00-00 00:00:00'), OrmComparator(cmp.item, 'ISNULL'))
            elif session.dbms_type == 'postgresql':
                return OrmComparator(cmp.item, 'ISNULL')
        return cmp


class ColumnDateAbstract(ColumnAttributeAbstract):
    def __init__(self, **kwargs):
        super(ColumnDateAbstract, self).__init__(**kwargs)
        self.data_type = 'date'
        self.value_class = AttributeValueDate
        self.strict_find = self.strict_find if self.strict_find is not None else True
        self.full_format = bool(kwargs.pop('full_format', False))
        self.format = kwargs.pop('format', None)

    def is_eq_with_tbl_schema(self, struct, session):
        if session in ('mysql', 'postgresql'):
            if str(struct['data_type']).lower() != 'date':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        d.append("DATE")
        if not self.nullable:
            d.append('NOT NULL')
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            d.append("SET DATA TYPE DATE")
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def text_repr(self, value, instance=None):
        if value is None:
            return ""
        session = manager.get_session(instance=instance)
        if isinstance(value, datetime.date) and not isinstance(value, datetime.datetime):
            format_ = self.format or \
                      (manager.get_full_date_format(session=session)
                       if self.full_format
                       else manager.get_short_date_format(session=session))
        elif isinstance(value, datetime.time):
            format_ = self.format or \
                      (manager.get_full_time_format(session=session)
                       if self.full_format
                       else manager.get_short_time_format(session=session))
        else:
            dateformat_ = self.format or \
                          (manager.get_full_date_format(session=session)
                           if self.full_format
                           else manager.get_short_date_format(session=session))
            timeformat_ = self.format or \
                          (manager.get_full_time_format(session=session)
                           if self.full_format
                           else manager.get_short_time_format(session=session))
            format_ = dateformat_ + " " + timeformat_
        return format_datetime(value, format_)

    @staticmethod
    def _verify_fail():
        raise OrmValueVerificationError(gettext("Must be a valid date"))

    def verify(self, value):
        value = super(ColumnDateAbstract, self).verify(value)
        if value is None:
            return None
        if isinstance(value, str):
            ps = value.split(' ')
            d_ = False
            n_year = 0
            n_month = 0
            n_day = 0
            for p in ps:
                if not p:
                    continue
                if '.' not in p and '/' not in p and '\\' not in p and '-' not in p and ':' not in p:
                    self._verify_fail()
                if ':' in p:
                    self._verify_fail()
                else:
                    if d_:
                        self._verify_fail()
                    s = '.' if '.' in p else ('-' if '-' in p else ('/' if '/' in p else '\\'))
                    p_ = p.split(s)
                    if len(p_) != 3:
                        self._verify_fail()
                    if not p_[0].isdigit() or not p_[1].isdigit() or not p_[2].isdigit():
                        self._verify_fail()
                    p1 = int(p_[0])
                    p2 = int(p_[1])
                    p3 = int(p_[2])
                    starting_with_year = p1 > 31 or s == '-'
                    if starting_with_year:
                        n_year = p1
                        if s in ('.', '-'):
                            n_month = p2
                            n_day = p3
                        else:
                            n_month = p3
                            n_day = p2
                    else:
                        n_year = p3
                        if s in ('.', '-'):
                            n_month = p2
                            n_day = p1
                        else:
                            n_month = p1
                            n_day = p2
                    if 99 < n_year < 1000:
                        self._verify_fail()
                    if n_year < 1000:
                        n_year = n_year + 1000 if n_year > 69 else n_year + 2000
                    if n_day == 0 or n_month == 0 or n_day > 31 or n_month > 12:
                        self._verify_fail()
                    n_daysmax = calendar.monthrange(n_year, n_month)[1]
                    if n_day > n_daysmax:
                        self._verify_fail()
                    d_ = True
            if not d_:
                self._verify_fail()
            value = datetime.date(n_year, n_month, n_day)
        elif not isinstance(value, datetime.date):
            self._verify_fail()
        return value

    @property
    def empty_(self):
        return OrmComparator(self, 'FALSE')

    @property
    def notempty_(self):
        return OrmComparator(self, 'TRUE')

    @property
    def true_(self):
        return OrmComparator(self, 'TRUE')

    @property
    def false_(self):
        return OrmComparator(self, 'FALSE')

    def resolve_comparator_(self, cmp, session):
        from ..funcs import or_, and_
        if cmp.cmp == 'TRUE':
            if session.dbms_type == 'mysql':
                return and_(OrmComparator(cmp.item, '!=', '0000-00-00'), OrmComparator(cmp.item, '!ISNULL'))
            elif session.dbms_type == 'postgresql':
                return OrmComparator(cmp.item, '!ISNULL')
        elif cmp.cmp == 'FALSE':
            if session.dbms_type == 'mysql':
                return or_(OrmComparator(cmp.item, '=', '0000-00-00'), OrmComparator(cmp.item, 'ISNULL'))
            elif session.dbms_type == 'postgresql':
                return OrmComparator(cmp.item, 'ISNULL')
        return cmp


class ColumnTimeAbstract(ColumnAttributeAbstract):
    def __init__(self, fsp=None, **kwargs):
        self.fsp = fsp
        super(ColumnTimeAbstract, self).__init__(**kwargs)
        self.data_type = 'time'
        self.value_class = AttributeValueTime
        self.strict_find = self.strict_find if self.strict_find is not None else True
        self.full_format = bool(kwargs.pop('full_format', False))
        self.format = kwargs.pop('format', None)
        self.use24hours = bool(kwargs.pop('use24hours', None))

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() != 'time':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if int(struct['datetime_precision']) != (int(self.fsp) if self.fsp is not None else 0):
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['column_type']).lower() != 'time':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if int(struct['datetime_precision']) != (int(self.fsp) if self.fsp is not None else 6):
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        d.append("TIME" if not self.fsp else "TIME(%i)" % int(self.fsp))
        if not self.nullable:
            d.append('NOT NULL')
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            d.append("SET DATA TYPE TIME")
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def text_repr(self, value, instance=None):
        if value is None:
            return ""
        session = manager.get_session(instance=instance)
        if isinstance(value, datetime.date) and not isinstance(value, datetime.datetime):
            format_ = self.format or \
                      (manager.get_full_date_format(session=session)
                       if self.full_format
                       else manager.get_short_date_format(session=session))
        elif isinstance(value, datetime.time):
            format_ = self.format or \
                      (manager.get_full_time_format(session=session)
                       if self.full_format
                       else manager.get_short_time_format(session=session))
        else:
            dateformat_ = self.format or \
                          (manager.get_full_date_format(session=session)
                           if self.full_format
                           else manager.get_short_date_format(session=session))
            timeformat_ = self.format or \
                          (manager.get_full_time_format(session=session)
                           if self.full_format
                           else manager.get_short_time_format(session=session))
            format_ = dateformat_ + " " + timeformat_
        return format_datetime(value, format_)

    @staticmethod
    def _verify_fail():
        raise OrmValueVerificationError(gettext("Must be a valid time"))

    def verify(self, value):
        value = super(ColumnTimeAbstract, self).verify(value)
        if value is None:
            return None
        if isinstance(value, str):
            ps = value.split(' ')
            t_ = False
            n_hour = 0
            n_minute = 0
            n_second = 0
            for p in ps:
                if not p:
                    continue
                if '.' not in p and '/' not in p and '\\' not in p and '-' not in p and ':' not in p:
                    self._verify_fail()
                if ':' in p:
                    if t_:
                        self._verify_fail()
                    if p.count(':') > 2:
                        self._verify_fail()
                    p_ = p.split(':')
                    if not p_[0].isdigit() or not p_[1].isdigit() or (len(p_) == 3 and not p_[2].isdigit()):
                        self._verify_fail()
                    n_hour = int(p_[0])
                    n_minute = int(p_[1])
                    n_second = int(p_[2]) if len(p) == 3 else 0
                    if n_hour > 23 or n_minute > 59 or n_second > 59:
                        self._verify_fail()
                    t_ = True
                else:
                    self._verify_fail()
            if not t_:
                self._verify_fail()
            value = datetime.time(n_hour, n_minute, n_second)
        elif not isinstance(value, datetime.time):
            self._verify_fail()
        return value

    @property
    def empty_(self):
        return OrmComparator(self, 'FALSE')

    @property
    def notempty_(self):
        return OrmComparator(self, 'TRUE')

    @property
    def true_(self):
        return OrmComparator(self, 'TRUE')

    @property
    def false_(self):
        return OrmComparator(self, 'FALSE')

    def resolve_comparator_(self, cmp, session):
        if cmp.cmp == 'TRUE':
            return OrmComparator(cmp.item, '!ISNULL')
        elif cmp.cmp == 'FALSE':
            return OrmComparator(cmp.item, 'ISNULL')
        return cmp


class ColumnStringAbstract(ColumnAttributeAbstract):
    def __init__(self, length=255, **kwargs):
        if length > 65535:
            raise OrmDefinitionError("string attribute types cannot be longer than 65535 symbols length!")
        elif length < 1:
            raise OrmDefinitionError("string attribute types' must be at least one symbol length!")
        self.length = length
        self.collate_ci = kwargs.pop('collate_ci', False)
        self.max_chars = kwargs.pop('max_chars', None)
        if self.max_chars is not None and not isinstance(self.max_chars, int):
            raise TypeError("'max_chars' must be None or (int) type!")
        super(ColumnStringAbstract, self).__init__(**kwargs)
        self.data_type = 'str'
        self.value_class = AttributeValueString
        if not self.nullable and not self._default:
            self._default = ''
        self.override_null = ''
        self.strict_find = self.strict_find if self.strict_find is not None else False

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() != 'varchar':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if struct['character_maximum_length'] != self.length:
                return False
            if not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and str(self._default) != str(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['column_type']).lower() != 'varchar':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            pg_attr_default = None \
                if self.default is None \
                else "'%s'::character varying" % str(self.default)
            if not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and pg_attr_default != str(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
            if struct['character_maximum_length'] != self.length:
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        d.append("VARCHAR(%i)" % self.length)
        if session.dbms_type == 'mysql':
            d.append("CHARACTER SET utf8 COLLATE utf8_bin")
        if not self.nullable:
            d.append('NOT NULL')
        if not callable(self._default) and self._default is not None:
            d.append("DEFAULT '%s'" % str(self._default).replace('\\', '\\\\').replace("'", "\'").replace("\n", "\\n"))
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            d.append("SET DATA TYPE VARCHAR(%i)" % self.length)
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            if not callable(self._default) and self._default is not None:
                d.append("SET DEFAULT '%s'"
                         % str(self._default).replace('\\', '\\\\').replace("'", "\'").replace("\n", "\\n"))
            else:
                d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def verify(self, value):
        value = super(ColumnStringAbstract, self).verify(value)
        if value is None:
            return None
        if self.max_chars and strlen(value) > int(self.max_chars):
            raise OrmValueVerificationError(gettext("Must be not longer than %i characters" % self.max_chars))
        return str(value)

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))

    @property
    def notempty_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def true_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def false_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))

    def contains_(self, value):
        return self.like_("%%%s%%" % value)

    def not_contains_(self, value):
        return self.notlike_("%%%s%%" % value)


class ColumnTextAbstract(ColumnAttributeAbstract):
    def __init__(self, **kwargs):
        self.max_chars = kwargs.pop('max_chars', None)
        if self.max_chars is not None and not isinstance(self.max_chars, int):
            raise TypeError("'max_chars' must be None or (int) type!")
        self.big = kwargs.pop('big', False)
        super(ColumnTextAbstract, self).__init__(**kwargs)
        self.data_type = 'text'
        self.value_class = AttributeValueString
        self.strict_find = self.strict_find if self.strict_find is not None else False

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() != ('longtext' if self.big else 'text'):
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['data_type']) != 'text':
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        if session.dbms_type == 'mysql':
            d.append("TEXT" if not self.big else "LONGTEXT")
        else:
            d.append("TEXT")
        if session.dbms_type == 'mysql':
            d.append("CHARACTER SET utf8 COLLATE utf8_bin")
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            d.append("SET DATA TYPE TEXG")
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def verify(self, value):
        value = super(ColumnTextAbstract, self).verify(value)
        if value is None:
            return None
        if self.max_chars and strlen(value) > int(self.max_chars):
            raise OrmValueVerificationError(gettext("Must be not longer than %i characters" % self.max_chars))
        return str(value)

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))

    @property
    def notempty_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def true_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def false_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))

    def contains_(self, value):
        return self.like_("%%%s%%" % value)

    def not_contains_(self, value):
        return self.notlike_("%%%s%%" % value)


class ColumnBlobAbstract(ColumnAttributeAbstract):
    def __init__(self, **kwargs):
        self.big = kwargs.pop('big', False)
        super(ColumnBlobAbstract, self).__init__(**kwargs)
        self.data_type = 'blob'
        self.value_class = AttributeValueBinary
        self.strict_find = self.strict_find if self.strict_find is not None else False

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() != ('longblob' if self.big else 'blob'):
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['data_type']).lower() != 'bytea':
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        if session.dbms_type == 'mysql':
            d.append("BLOB" if not self.big else "LONGBLOB")
        else:
            d.append("BYTEA")
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            d.append("SET DATA TYPE BYTEA")
            d.append("DROP NOT NULL")
            d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))

    @property
    def notempty_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def true_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def false_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))


class ColumnBooleanAbstract(ColumnAttributeAbstract):
    def __init__(self, **kwargs):
        self.boolean = True
        if kwargs.get('is_softdelete_key', False):
            kwargs['nullable'] = kwargs.get('nullable', False)
            kwargs['default'] = kwargs.get('default', False)
        super(ColumnBooleanAbstract, self).__init__(**kwargs)
        self.override_null = False
        self.data_type = 'bool'
        self.value_class = AttributeValueBoolean

    def is_eq_with_tbl_schema(self, struct, session):
        if session.dbms_type == 'mysql':
            if str(struct['data_type']).lower() != 'tinyint':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            if not callable(self._default):
                if self._default is None and struct['column_default'] is not None:
                    return False
                elif self._default is not None and struct['column_default'] is None:
                    return False
                elif self._default is not None and bool(self._default) != bool(struct['column_default']):
                    return False
            elif struct['column_default'] is not None:
                return False
        elif session.dbms_type == 'postgresql':
            if str(struct['data_type']).lower() != 'boolean':
                return False
            if struct['is_nullable'] != self.nullable:
                return False
            attr_default = None \
                if callable(self._default) \
                else (bool(self._default) if self._default is not None else None)
            column_default = None \
                if struct['column_default'] is None \
                else (True if str(struct['column_default']).lower() in ('true', 'y', 'yes') else False)
            if not callable(self._default):
                if attr_default is None and column_default is not None:
                    return False
                elif attr_default is not None and column_default is None:
                    return False
                elif attr_default is not None and attr_default != column_default:
                    return False
            elif struct['column_default'] is not None:
                return False
        return True

    def declaration_sql(self, session):
        d = list()
        d.append("BOOL")
        if not self.nullable and not self.auto_increment:
            d.append('NOT NULL')
        if not callable(self._default) and self._default is not None:
            d.append('DEFAULT TRUE' if bool(self._default) else 'DEFAULT FALSE')
        return " ".join(d)

    def alter_sql(self, session):
        if session.dbms_type == 'postgresql':
            d = list()
            d.append("SET DATA TYPE BOOL")
            d.append("DROP NOT NULL" if self.nullable else "SET NOT NULL")
            if not callable(self._default) and self._default is not None:
                d.append("SET DEFAULT %s" % ("TRUE" if bool(self._default) else "FALSE"))
            else:
                d.append("DROP DEFAULT")
            return d
        return self.declaration_sql(session)

    def like_(self, value):
        raise OrmComparatorValueError("boolean attribute types cannot be filtered using `like_` or `notlike_`!")

    def notlike_(self, value):
        raise OrmComparatorValueError("boolean attribute types cannot be filtered using `like_` or `notlike_`!")

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, 'FALSE'))

    @property
    def notempty_(self):
        return OrmComparator(self, 'TRUE')




import copy
from pc23 import xrange
from ..exceptions import OrmParameterError, OrmAttributeTypeMistake, OrmInvalidKey, OrmDefinitionError, \
    OrmModelDeclarationError
from ..mapper import mapper
from ..manage import manager
from ..values.relationships import ValueRelatedToOne, ValueRelatedToMany, ValueManyToMany
from .base import OrmComparatorMixin, AttributeAliased, AttributeAbstract, KeyAttributeAbstract, \
    ColumnAttributeAbstract, DataAttributeAbstract, FK_ONDELETE_CASCADE, FK_ONDELETE_SET_NULL


class ForeignKeyConstraintAbstract(KeyAttributeAbstract):
    def __init__(self, *args, **kwargs):
        if len(args) != 2:
            raise OrmParameterError(
                "Relationships requires pairs of _attrs to be specified! For complex relationships, when"
                " more than one column from each of OrmModel participates, they must be grouped using (list)"
                " or (tuple)."
            )
        if (isinstance(args[0], (list, tuple)) and not isinstance(args[1], (list, tuple))) \
                or (isinstance(args[1], (list, tuple)) and not isinstance(args[0], (list, tuple))):
            raise OrmParameterError(
                "Complex relationships cannot be declared with different quantity of _attrs for"
                " one and for another Models! Local and referenced _attrs must be declared by pairs!"
            )
        if isinstance(args[0], (list, tuple)) and len(args[0]) != len(args[1]):
            raise OrmParameterError(
                "Complex relationships cannot be declared with different quantity of _attrs for"
                " one and for another Models! Local and referenced _attrs must be declared by pairs!"
            )
        super(ForeignKeyConstraintAbstract, self).__init__(*args)
        self.on_delete = kwargs.pop('on_delete', None)
        self.on_update = kwargs.pop('on_update', None)
        self.ref_model = None
        self.ref_columns = list()

    def validate(self):
        if self._validated:
            return
        local_args = self._args[0] if isinstance(self._args[0], (list, tuple)) else (self._args[0], )
        ref_args = self._args[1] if isinstance(self._args[1], (list, tuple)) else (self._args[1],)
        ref_model_name = None
        for arg in local_args:
            if not isinstance(arg, (str, ColumnAttributeAbstract)):
                raise OrmParameterError(
                    "Columns in relationship declarations must be set using textual names (str) or attributes"
                    " themselfs."
                )
            if isinstance(arg, str) and '.' in arg:
                raise OrmAttributeTypeMistake(
                    "Local (for OrmModel) _attrs, defined for relationships, cannot be specified with"
                    " OrmModel name part ('OrmModel.attribute_name'), but must be specified using only attribute"
                    " name ('attribute_name')!"
                )
            k = arg if isinstance(arg, str) else arg.model_k
            if not hasattr(self.model, k):
                raise OrmInvalidKey(
                    "OrmModel '%s' has no attribute '%s' set for the relationship '%s'!" %
                    (self.model.__name__, k, self.model_k)
                )
            c = getattr(self.model, k)
            if not isinstance(c, ColumnAttributeAbstract):
                raise OrmAttributeTypeMistake(
                    "Attribute '%s' in the OrmModel '%s', declaring in the key '%s', is not a type of"
                    " Column! Only database table refrected attributes may be set for keys!" %
                    (k, self.model.__name__, self.model_k)
                )
            self.columns.append(k)
        for arg in ref_args:
            if not isinstance(arg, (str, ColumnAttributeAbstract)):
                raise OrmParameterError(
                    "Columns in relationship declarations must be set using textual names (str) or attributes"
                    " themselfs."
                )
            if isinstance(arg, str) and '.' not in arg:
                raise OrmAttributeTypeMistake(
                    "Referencing _attrs, defined for relationships, must be specified with"
                    " OrmModel name part ('OrmModel.attribute_name')!"
                )
            if isinstance(arg, str):
                p = arg.split('.')
                if len(p) != 2:
                    raise OrmAttributeTypeMistake(
                        "Referencing _attrs, defined for relationships, must be specified with"
                        " OrmModel name part ('OrmModel.attribute_name')!"
                    )
                attr_ref_model = p[0]
                if ref_model_name and ref_model_name != attr_ref_model:
                    raise OrmParameterError("There can be only one referencing OrmModel for relationship!")
                elif not ref_model_name:
                    ref_model_name = attr_ref_model
                    self.ref_model = mapper.get_model_by_name(ref_model_name)
                k = p[1]
                if not hasattr(self.ref_model, k):
                    raise OrmInvalidKey(
                        "OrmModel `%s` has no attribute `%s` set for the relationship `%s`!" %
                        (self.ref_model.__name__, k, self.model_k)
                    )
                c = getattr(self.ref_model, k)
                if not isinstance(c, ColumnAttributeAbstract):
                    raise OrmAttributeTypeMistake(
                        "Attribute `%s` in the OrmModel `%s`, declaring in the key `%s`, is not a type of"
                        " Column! Only database table refrected attributes may be set for keys!" %
                        (k, self.model.__name__, self.model_k)
                    )
            else:
                attr_ref_model = arg.model.__name__
                if ref_model_name and ref_model_name != attr_ref_model:
                    raise OrmParameterError("There can be only one referencing OrmModel for relationship!")
                elif not ref_model_name:
                    ref_model_name = attr_ref_model
                    self.ref_model = mapper.get_model_by_name(ref_model_name)
                k = arg.model_k
                if not isinstance(arg, ColumnAttributeAbstract):
                    raise OrmAttributeTypeMistake(
                        "Attribute `%s` in the OrmModel `%s`, declaring in the key `%s`, is not a type of"
                        " Column! Only database table refrected attributes may be set for keys!" %
                        (k, self.model.__name__, self.model_k)
                    )
            self.ref_columns.append(k)
        self._validated = True

    @property
    def key_name(self):
        return "%s_ibfk_%s" % (self.model.table, self.tbl_k)


class ForeignKey(ForeignKeyConstraintAbstract):
    """ Foreign key constraint which refrecting at the database table. """
    def declaration_sql(self, session):
        self.validate()
        fk_name = self.tbl_k
        fk_l = []
        fk_r = []
        fk_sz = len(self.columns)
        for i in xrange(fk_sz):
            fk_l_ = self.columns[i]
            fk_r_ = self.ref_columns[i]
            fk_cl_ = getattr(self.model, fk_l_)
            fk_cr_ = getattr(self.ref_model, fk_r_)
            fk_l.append(session.quote_name(fk_cl_.tbl_k))
            fk_r.append(session.quote_name(fk_cr_.tbl_k))
        fk_table = self.ref_model.table
        fk_left = ",".join(fk_l)
        fk_right = ",".join(fk_r)
        fkq = "CONSTRAINT %s FOREIGN KEY (%s) REFERENCES %s(%s)" % \
              (session.quote_name(fk_name), fk_left, session.quote_name(fk_table), fk_right)
        if self.on_delete:
            fkq += " ON DELETE " + str(self.on_delete).upper()
        if self.on_update:
            fkq += " ON UPDATE " + str(self.on_update).upper()
        return fkq

    def declaration_key_sql(self, session):
        self.validate()
        fk_name = self.tbl_k
        fk_l = []
        fk_sz = len(self.columns)
        for i in xrange(fk_sz):
            fk_l_ = self.columns[i]
            fk_cl_ = getattr(self.model, fk_l_)
            fk_l.append(session.quote_name(fk_cl_.tbl_k))
        fk_left = ",".join(fk_l)
        return "KEY %s (%s)" % (session.quote_name(fk_name), fk_left)


class VirtualForeignKey(ForeignKeyConstraintAbstract):
    """ Virtual relationship which is actual for ORM mechanics same as foreign keys, but
    not reflecting at the database table. """
    pass


class Relationship(AttributeAbstract):
    """ A special kind of attribute which make a relationship between two Models - the one
    where it is declared at, and the second which is specified as a first parameter. This
    attribute analyses the destination OrmModel and creates local attribute with the same
    type and secondary parameters as at destination OrmModel, so the programmer have not to
    care about correct type of local attribute (which must be compatible with foreign
    attribute at foreign OrmModel otherwise). Besides, this attribute automatically declare
    a foreign key in the local OrmModel, connecting it with foreign OrmModel using local
    attribute.
    A destination (foreign side) can be set using:
    a) a textual name of the attribute in the format 'foreign_Model_name.attribute_name';
    b) a testual name of the foreign OrmModel 'foreign_Model_name';
    c) an attribute itself: `ForeignModel.attribute_name`;
    d) a foreign OrmModel itself: `ForeignModel`;
    When no foreign attribute is set - ORM will take a primary key of the foreign
    OrmModel and use it as foreign attribute.
    """
    def __init__(self, foreign_attribute, on_delete=None, on_update=None, **kwargs):
        super(Relationship, self).__init__()
        self.on_delete = on_delete
        self.on_update = on_update
        self.foreign_attribute = foreign_attribute
        self.kw = kwargs

    @staticmethod
    def _deferred_make(local_model, local_k, foreign_model, foreign_k, **kwargs):
        local_model = mapper.get_model_by_name(local_model)
        foreign_model = mapper.get_model_by_name(foreign_model)
        if foreign_k is None:
            foreign_pks = foreign_model.__meta__.primary_key
            if len(foreign_pks) != 1:
                raise OrmParameterError(
                    "(Relationship) can make relationship using only simple pair of local and"
                    " foreign attributes, but given foreign OrmModel has complex primary key defined!"
                )
            foreign_k = foreign_pks[0]
        local_c = getattr(local_model, local_k)
        foreign_c = getattr(foreign_model, foreign_k)
        if not isinstance(foreign_c, ColumnAttributeAbstract):
            raise OrmParameterError(
                "(Relationship) type attribute requries that foreign attribute which this pointer is"
                " pointing at to be a type of: a) (str) in format 'Model_name.attribute_name';"
                " b) (str) in format 'Model_name' - then OrmModel's primary key will be taken as"
                " a destination attribute; c) an attribute itself or d) a OrmModel - then OrmModel's"
                " primary key will be taken as a destination attribute!"
            )
        fk_type = VirtualForeignKey if isinstance(local_c, VirtualRelationship) else ForeignKey
        on_delete = local_c.on_delete
        on_update = local_c.on_update
        foreign_c_type = foreign_c.base_type or foreign_c.__class__
        override_c = foreign_c_type()
        override_c.__dict__ = copy.deepcopy(foreign_c.__dict__)
        override_c.primary_key = False
        override_c.auto_increment = False
        override_c.tbl_k = local_c.tbl_k
        override_c.declaration_order = local_c.declaration_order
        override_c.nullable = kwargs.pop('nullable', True)
        if foreign_c.primary_key:
            override_c.default = None
        kwargs.pop('auto_increment', None)
        kwargs.pop('primary_key', None)
        kwargs.pop('tbl_k', None)
        for k in kwargs:
            setattr(override_c, k, kwargs[k])
        setattr(local_model, local_k, override_c)
        fk_c = fk_type(override_c, foreign_c, on_delete=on_delete, on_update=on_update)
        fk_k = "fk_%s_" % local_k
        setattr(local_model, fk_k, fk_c)
        override_c.foreign_key = fk_c

    def on_declare(self):
        from ..models.meta import OrmModelMeta
        from ..models.base import OrmModelAbstract
        if isinstance(self.foreign_attribute, str):
            p = self.foreign_attribute.split('.')
            if len(p) > 2:
                raise OrmParameterError(
                    "(Relationship) type attribute requries that foreign attribute which this pointer is"
                    " pointing at to be a type of: a) (str) in format 'Model_name.attribute_name';"
                    " b) (str) in format 'Model_name' - then OrmModel's primary key will be taken as"
                    " a destination attribute; c) an attribute itself or d) a OrmModel - then OrmModel's"
                    " primary key will be taken as a destination attribute!"
                )
            foreign_model_name = p[0]
            foreign_k = p[1] if len(p) == 2 else None
        elif isinstance(self.foreign_attribute, ColumnAttributeAbstract):
            foreign_model_name = self.foreign_attribute.model.__name__
            foreign_k = self.foreign_attribute.model_k
        elif isinstance(self.foreign_attribute, (OrmModelMeta, OrmModelAbstract)):
            foreign_model_name = self.foreign_attribute.__name__
            foreign_k = None
        else:
            raise OrmParameterError(
                "(Relationship) type attribute required that first argument to be a name of foreign"
                " attribute in format 'OrmModel.attribute_name' or a foreign attribute itself!"
            )
        args = (self.model.__name__, self.model_k, foreign_model_name, foreign_k)
        kwargs = self.kw
        mapper.register_on_model_declare(foreign_model_name, self._deferred_make, *args, **kwargs)


class VirtualRelationship(Relationship):
    pass


class RelatedAbstract(AttributeAbstract):
    def __init__(self, foreign, on_delete=None, **kwargs):
        super(RelatedAbstract, self).__init__()
        self._validated = False
        self.foreign = foreign
        self.foreign_model = None
        self.foreign_k = None
        self.foreign_c = None
        self.on_delete = on_delete
        self._caption = kwargs.pop('caption', None)
        self._abbreviation = kwargs.pop('abbreviation', None)

    @property
    def caption(self):
        """ The textual full-sized caption of the attribute. Not used at DBMS side. """
        return self._get_f_value('_caption', None) or self._get_f_value('_abbreviation', None) or self.model_k

    @caption.setter
    def caption(self, value):
        """ Sets the caption of this attribute. """
        self._caption = value

    @property
    def abbreviation(self):
        """ The abbreviation of the attribute. Not used at DBMS side. """
        return self._get_f_value('_abbreviation', None) or self._get_f_value('_caption', None) or self.model_k

    @abbreviation.setter
    def abbreviation(self, value):
        """ Sets the abbreviation of the attribute. """
        self._abbreviation = value

    def on_model_declared(self):
        if isinstance(self.foreign, AttributeAbstract):
            self.foreign = "%s.%s" % (self.foreign.model.__name__, self.foreign.model_k)

    def validate(self):
        from ..models.meta import OrmModelMeta
        from ..models.base import OrmModelAbstract
        if self._validated:
            return
        if isinstance(self.foreign, str):
            if '.' in self.foreign:
                p = self.foreign.split('.')
                if len(p) != 2:
                    raise OrmParameterError(
                        "Related pointers must be set using OrmModel wich has corresponding relationship with"
                        " this model declared, or using format `OrmModel.key` where `key` can be a relationship or"
                        " foreign key describing the relationship between foreign model and this one!"
                    )
                ref_model_name_ = p[0]
                ref_k_ = p[1]
                if not ref_model_name_ or not ref_k_:
                    raise OrmParameterError(
                        "Related pointers must be set using OrmModel wich has corresponding relationship with"
                        " this model declared, or using format `OrmModel.key` where `key` can be a relationship or"
                        " foreign key describing the relationship between foreign model and this one!"
                    )
                ref_model_ = mapper.get_model_by_name(ref_model_name_)
                if not ref_model_:
                    raise OrmParameterError(
                        "Related pointer referencing unknown OrmModel name: `%s`!" % ref_model_name_
                    )
                if not hasattr(ref_model_, ref_k_):
                    raise OrmParameterError(
                        "Related pointer referencing unknown OrmModel's attribute: `%s`.`%s`!" %
                        (ref_model_name_, ref_k_)
                    )
                foreign_c_ = getattr(ref_model_, ref_k_)
                if getattr(foreign_c_, 'foreign_key', None):
                    foreign_c_ = foreign_c_.foreign_key
                foreign_c_.validate()
                foreign_model_ = foreign_c_.ref_model if foreign_c_.ref_model != self.model else foreign_c_.model
            elif mapper.has_model(self.foreign):
                foreign_model_name_ = self.foreign
                foreign_model_ = mapper.get_model_by_name(foreign_model_name_)
                if not foreign_model_:
                    raise OrmParameterError(
                        "Related pointer referencing unknown OrmModel name: `%s`!" % foreign_model_name_
                    )
                foreign_c_ = foreign_model_.__meta__.get_relationship_with_model(self.model)
                if foreign_c_ is None and not foreign_model_.__meta__.foreign:
                    foreign_c_ = foreign_model_.__meta__.make_relationship_with_model(
                        self.model, on_delete=self.on_delete)
                if foreign_c_ is None:
                    raise OrmParameterError(
                        "Related pointer referencing OrmModel `%s` which has no declared relationship key with"
                        " this OrmModel!" % foreign_model_name_
                    )
                elif foreign_c_ is True:
                    raise OrmParameterError(
                        "Related pointer referencing OrmModel `%s` which has more than one unique relationship"
                        " with this OrmModel! Is this case there must be a referencing attribute to be specified"
                        " using format `OrmModel.attribute`!"
                    )
                foreign_c_.validate()
            else:
                ref_k_ = self.foreign
                if not hasattr(self.model, ref_k_):
                    raise OrmParameterError(
                        "Related pointer referencing OrmModel `%s` which has more than one unique relationship"
                        " with this OrmModel! Is this case there must be a referencing attribute to be specified"
                        " using format `OrmModel.attribute`!"
                    )
                foreign_c_ = getattr(self.model, ref_k_)
                if getattr(foreign_c_, 'foreign_key', None):
                    foreign_c_ = foreign_c_.foreign_key
                foreign_c_.validate()
                foreign_model_ = foreign_c_.ref_model
        elif isinstance(self.foreign, (OrmModelMeta, OrmModelAbstract)):
            foreign_model_name_ = self.foreign.__class__.__name__
            foreign_model_ = self.foreign
            foreign_c_ = foreign_model_.__meta__.get_relationship_with_model(self.model)
            if foreign_c_ is None:
                raise OrmParameterError(
                    "Related pointer referencing OrmModel `%s` which has no declared relationship key with"
                    " this OrmModel!" % foreign_model_name_
                )
            elif foreign_c_ is True:
                raise OrmParameterError(
                    "Related pointer referencing OrmModel `%s` which has more than one unique relationship"
                    " with this OrmModel! Is this case there must be a referencing attribute to be specified"
                    " using format `OrmModel.attribute`!"
                )
            foreign_c_.validate()
        elif isinstance(self.foreign, ForeignKeyConstraintAbstract):
            foreign_model_ = self.foreign.model
            foreign_c_ = self.foreign
            foreign_c_.validate()
        else:
            raise OrmParameterError(
                "Related pointers must be set using OrmModel wich has corresponding relationship with this"
                " model declared, or using format `OrmModel.key` where `key` can be a relationship or"
                " foreign key describing the relationship between foreign model and this one!"
            )
        if getattr(foreign_c_, 'foreign_key', None):
            foreign_c_ = foreign_c_.foreign_key
        self.foreign_model = foreign_model_
        self.foreign_c = foreign_c_
        self.foreign_c.validate()
        self._validated = True


class RelatedToOneAbstract(RelatedAbstract):
    def __init__(self, foreign, on_delete=None, always_deferred=False, **kwargs):
        super(RelatedToOneAbstract, self).__init__(foreign, on_delete=on_delete, **kwargs)
        self.always_deferred = bool(always_deferred)

    def initialize(self, instance, k):
        virtual_value = ValueRelatedToOne()
        virtual_value.model_k = k
        virtual_value.instance = instance
        virtual_value.c = self
        instance.__setattrabs__(k, virtual_value)


class RelatedToManyAbstract(RelatedAbstract):
    def initialize(self, instance, k):
        virtual_value = ValueRelatedToMany()
        virtual_value.model_k = k
        virtual_value.instance = instance
        virtual_value.c = self
        instance.__setattrabs__(k, virtual_value)


class OneToOne(RelatedToOneAbstract):
    pass


class ManyToOne(RelatedToOneAbstract):
    pass


class OneToMany(RelatedToManyAbstract):
    pass


class UserIdentAbstract(ManyToOne):
    def __init__(self, **kwargs):
        self.ident_k = None
        super(UserIdentAbstract, self).__init__(None, on_delete=FK_ONDELETE_SET_NULL, **kwargs)

    def on_model_declared(self):
        foreign = manager.get_user_ident()
        fmodel = mapper.get_model_by_name(foreign) if isinstance(foreign, str) else foreign
        fpk = fmodel.__meta__.get_primary_key()
        if len(fpk) > 1:
            raise OrmModelDeclarationError("user ident's Model must not have complex primary key!")
        fpk = fpk[0]
        foreign = "%s.%s" % (fmodel.__name__, fpk.model_k)
        c = Relationship(foreign, on_delete=FK_ONDELETE_SET_NULL)
        c.declaration_order = 15
        k = "%s_id" % self.model_k
        setattr(self.model, k, c)
        self.foreign = k
        self.ident_k = k
        super(UserIdentAbstract, self).on_model_declared()

    def validate(self):
        super(UserIdentAbstract, self).validate()


class AuthorIdent(UserIdentAbstract):
    """ This attribute will be set to the current session user ID/UUID when the object is a new one
    and only inserting to the database - the initial author of the object. """
    def on_model_declared(self):
        super(AuthorIdent, self).on_model_declared()
        self.model.__meta__.author_ident.append(self)


class PerformerIdent(UserIdentAbstract):
    """ This attribute will be set to the current session user ID/UUID always when object writes,
    not matter - is the object new or not. """
    def on_model_declared(self):
        super(PerformerIdent, self).on_model_declared()
        self.model.__meta__.performer_ident.append(self)


class UserIdent(UserIdentAbstract):
    """ The custom user ident attribute - it will not automatically be set to anything. """
    pass


class ManyToMany(AttributeAbstract):
    def __init__(self, model_a, model_b, on_delete=FK_ONDELETE_CASCADE):
        from ..models.meta import OrmModelMeta
        from ..models.base import OrmModel
        if not isinstance(model_a, (OrmModelMeta, OrmModel, str)) \
                or not isinstance(model_b, (OrmModelMeta, OrmModel, str)):
            raise OrmParameterError(
                "`ManyToMany` requires that the first and second arguments be of type (OrmModel|str)!"
            )
        if on_delete == FK_ONDELETE_SET_NULL:
            raise OrmParameterError("`ManyToMany`.`on_delete` cannot be `SET NULL`!")
        super(ManyToMany, self).__init__()
        self.model_a = model_a.__name__ if isinstance(model_a, (OrmModelMeta, OrmModel)) else model_a
        self.model_b = model_b.__name__ if isinstance(model_b, (OrmModelMeta, OrmModel)) else model_b
        self.on_delete = on_delete
        self.junction_model = None
        self.key = mapper.register_many2many(self, model_a, model_b)
        self._validated = False

    def validate(self, jmodel=None):
        if self._validated:
            return
        if self.key is None:
            return
        model = jmodel if jmodel is not None else mapper.get_many2many_junction_model(self.key)
        if not model:
            raise OrmDefinitionError(
                "`ManyToMany`.`validate` been called before Many-To-Many relationship initialized!"
            )
        attrs = dir(model)
        for k in attrs:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = getattr(model, k)
            if not isinstance(c, ForeignKeyConstraintAbstract):
                continue
            c.validate()
        self.junction_model = model
        self._validated = True

    def initialize(self, instance, k):
        virtual_value = ValueManyToMany()
        virtual_value.model_k = k
        virtual_value.instance = instance
        virtual_value.c = self
        instance.__setattrabs__(k, virtual_value)

    def is_assigned(self, f_instance):  # Abstract
        pass

    def find(self, *args, **kwargs):  # Abstract
        pass

    def add(self, *args, **kwargs):  # Abstract
        pass

    def delete(self, *args, **kwargs):  # Abstract
        pass

    def write(self, session=None, force=False):  # Abstract
        pass



class ForeignAttr(OrmComparatorMixin, DataAttributeAbstract):
    def __init__(self, related_attr, referenced_k=None, **kwargs):
        if not isinstance(related_attr, (str, RelatedToOneAbstract)):
            raise OrmParameterError(
                "ForeignColumn 'related_attr' parameter must be a type of (str) or ...ToOne relationship!"
            )
        self._related_attr = related_attr
        self._related_k = None
        self._ref_k = referenced_k
        super(ForeignAttr, self).__init__(**kwargs)
        self._custom_getter = self.getter
        self._custom_setter = self.setter
        self.getter = self._getter
        self.setter = self._setter

    def __hash__(self):
        return super(OrmComparatorMixin, self).__hash__()

    def as_(self, name):
        """ Returns this attribute formatted to be named column as `name`. """
        return AttributeAliased(self, name)

    @property
    def related_k(self):
        if self._related_k:
            return self._related_k
        if not isinstance(self._related_attr, (str, RelatedToOneAbstract)):
            raise OrmParameterError(
                "ForeignColumn 'related_attr' parameter must be a type of (str) or ...ToOne relationship!"
            )
        self._related_k = self._related_attr if isinstance(self._related_attr, str) else self._related_attr.model_k
        return self._related_k

    @property
    def referenced_k(self):
        return self._ref_k if self._ref_k else self.model_k

    def _getter(self, instance, k):
        if not hasattr(instance, self.related_k):
            raise OrmDefinitionError(
                "model '%s' has no relationship with name '%s' for foreign attribute '%s'" %
                (self.model.__name__, self._related_k, self.model_k)
            )
        ref_instance = getattr(instance, self.related_k)
        return getattr(ref_instance, self.referenced_k)

    def _setter(self, instance, k, value, on_instance_loaded=False):
        if not hasattr(instance, self.related_k):
            raise OrmDefinitionError(
                "model '%s' has no relationship with name '%s' for foreign attribute '%s'" %
                (self.model.__name__, self._related_k, self.model_k)
            )
        ref_instance = getattr(instance, self.related_k)
        return ref_instance.__setattr__(self.referenced_k, value)





import copy
import inspect
from pc23 import xrange
from ..mapper import mapper
from ..utils import gettext
from ..exceptions import OrmParameterError, OrmInvalidKey, OrmValueVerificationError, OrmAttributeTypeMistake, \
    OrmComparatorValueError
from ..values import AttributeValueNull


FK_ONDELETE_CASCADE = 'CASCADE'
FK_ONDELETE_SET_NULL = 'SET NULL'
FK_ONDELETE_RESTRICT = 'RESTRICT'
FK_ONDELETE_ON_ACTION = 'RESTRICT'
FK_ONUPDATE_CASCADE = 'CASCADE'
FK_ONUPDATE_SET_NULL = 'SET NULL'
FK_ONUPDATE_RESTRICT = 'RESTRICT'
FK_ONUPDATE_ON_ACTION = 'RESTRICT'


class OrmComparatorMixin(object):
    """ Abstract comparator class which may be inherited by other classes to
    give ORM condition processing, like equals to, greater than and etc. """
    def __eq__(self, other):
        if other is None:
            return OrmComparator(self, 'ISNULL')
        return OrmComparator(self, '=', other)

    def __ne__(self, other):
        if other is None:
            return OrmComparator(self, '!ISNULL')
        return OrmComparator(self, '!=', other)

    def __gt__(self, other):
        return OrmComparator(self, '>', other)

    def __lt__(self, other):
        return OrmComparator(self, '<', other)

    def __ge__(self, other):
        return OrmComparator(self, '>=', other)

    def __le__(self, other):
        return OrmComparator(self, '<=', other)

    def __invert__(self):
        return OrmComparator(self, '~')

    def __lshift__(self, other):
        return OrmComparator(self, '<<', other)

    def __rshift__(self, other):
        return OrmComparator(self, '>>', other)

    def __and__(self, other):
        return OrmComparator(self, '&', other)

    def __or__(self, other):
        return OrmComparator(self, '|', other)

    def __xor__(self, other):
        return OrmComparator(self, '^', other)

    def is_(self, other):
        if other is None:
            return OrmComparator(self, 'ISNULL')
        return self.__eq__(other)

    def isnot_(self, other):
        if other is None:
            return OrmComparator(self, '!ISNULL')
        return self.__ne__(other)

    @property
    def isnull_(self):
        return OrmComparator(self, 'ISNULL')

    @property
    def notnull_(self):
        return OrmComparator(self, '!ISNULL')

    @property
    def empty_(self):
        from ..funcs import or_
        return or_(OrmComparator(self, 'ISNULL'), OrmComparator(self, '=', ''))

    @property
    def notempty_(self):
        from ..funcs import and_
        return and_(OrmComparator(self, '!ISNULL'), OrmComparator(self, '!=', ''))

    @property
    def true_(self):
        return OrmComparator(self, 'TRUE')

    @property
    def false_(self):
        return OrmComparator(self, 'FALSE')

    def in_(self, *args):
        if len(args) == 0:
            raise OrmParameterError("Column.in_() requires at least one value to be given!")
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = list(args[0])
        return OrmComparator(self, 'IN', args)

    def notin_(self, *args):
        if len(args) == 0:
            raise OrmParameterError("Column.in_() requires at least one value to be given!")
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = list(args[0])
        return OrmComparator(self, '!IN', args)

    def like_(self, value):
        return OrmComparator(self, 'LIKE', value)

    def notlike_(self, value):
        return OrmComparator(self, '!LIKE', value)

    def contains_(self, value):
        return self.__eq__(value)

    def not_contains_(self, value):
        return self.__ne__(value)

    def find_(self, value):
        from ..funcs import lower_, textual_
        if not isinstance(value, str):
            if value is None:
                raise OrmComparatorValueError("cannot search for NULL value using 'find_' comparator!")
            value = str(value)
        return OrmComparator(lower_(textual_(self)), 'LIKE', "%%%s%%" % value.lower()) \
            if not self.strict_find \
            else OrmComparator(lower_(textual_(self)), '=', value.lower())

    def resolve_comparator_(self, cmp, session):
        return cmp

    def corresponds_to_(self, value):
        return OrmComparator(self.resolve_opts_to_case(), '=', value)

    def isas_(self, value):
        return self.corresponds_to_(value)


class OrmComparator(object):
    """ A class, used for specific filter (where, having,...) to be declared """
    def __init__(self, item, cmp, value=None):
        self.cmp = cmp
        self.item = item
        self.value = value

    def resolve(self, session):
        return self.item.resolve_comparator_(self, session)


class AttributeAbstract(object):
    """ An abstract class for all declared as data types and relationships attributes
    in the model. """
    declaration_order = 10

    def __new__(cls, *args, **kwargs):
        instance = super(AttributeAbstract, cls).__new__(cls)
        instance.declaration_order = AttributeAbstract.declaration_order
        instance.template_model = None
        instance.template_k = None
        AttributeAbstract.declaration_order += 10
        return instance

    def __init__(self):
        self._validated = True
        self.auto_created = False
        self.model = None
        self.tbl_k = None
        self.model_k = None
        self.base_type = None
        self.parent_model_k = None
        self.parent_tbl_k = None
        self.parent_model = None
        self.belongs_to_template = False
        self.template_model = None
        self.template_k = None
        self.on_new = None
        self.on_change = None
        self.setter = None
        self.getter = None
        self.foreign_key = None
        self.triggers = list()

    def __repr__(self):
        if self.model is not None:
            return "%s(%s.%s)" % (self.__class__.__name__, self.model.__name__, self.model_k)
        else:
            return str(self.__class__.__name__)

    def _attr_from_template(self, inst_value):
        from ..models.base import OrmModel
        if inspect.isclass(inst_value) and issubclass(inst_value, OrmModel) and inst_value == self.template_model:
            # Reassigning attribute's property pointing to the template itself - to the
            # end model itself.
            inst_value = self.model
            return inst_value
        if isinstance(inst_value, AttributeAbstract) and inst_value.belongs_to_template:
            # Reassigning a pointer to the attribute of the template to the corresponding
            # attribute of end Model.
            inst_value = getattr(self.model, inst_value.model_k, None)
            return inst_value
        if isinstance(inst_value, dict):
            for k, v in dict(inst_value).items():
                inst_value[k] = self._attr_from_template(v)
            return inst_value
        if isinstance(inst_value, list):
            for i in xrange(len(inst_value)):
                inst_value[i] = self._attr_from_template(inst_value[i])
            return inst_value
        if isinstance(inst_value, tuple):
            l_ = list()
            for i in inst_value:
                l_.append(self._attr_from_template(i))
            return tuple(l_)
        return inst_value

    def from_template(self):
        keys = dir(self)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            cls_attr = getattr(self.__class__, k, None)
            if type(cls_attr) in (property, bool, int, float, str) \
                    or hasattr(cls_attr, '__call__') \
                    or callable(cls_attr):
                continue
            inst_value = getattr(self, k)
            if inst_value is None or type(inst_value) in (bool, int, float, str):
                continue
            setattr(self, k, self._attr_from_template(inst_value))

    def get_instance_value(self, instance):
        return instance.__getattribute__(self.model_k)

    @classmethod
    def get_next_declaration_order(cls):
        r = cls.declaration_order
        cls.declaration_order += 10
        return r

    def _get_f_value(self, p_name, default=None):
        v = getattr(self, p_name)
        if v is None:
            return default
        if callable(v):
            return v()
        return v

    def on_declare(self):
        pass

    def on_model_declared(self):
        pass

    def validate(self):
        pass

    def initialize(self, instance, k):
        pass

    def register_trigger(self):
        pass

    def call_on_change(self, instance, key, value):
        if self.on_change is not None:
            if not callable(self.on_change):
                raise OrmAttributeTypeMistake(
                    "attribute's `on_change` must be a callable function with at least three arguments:"
                    " (instance, key, value)!"
                )
            self.on_change(instance, key, value)
        if self.triggers:
            for t in self.triggers:
                t(self, instance, key, value)

    def on_instance_load(self, instance):
        pass


class DataAttributeAbstract(AttributeAbstract):
    """ The type of model's attribute which is able to store some value itself (directly
    by holding a kind of value, or virtually) """
    def __init__(self, **kwargs):
        super(DataAttributeAbstract, self).__init__()
        self.virtual = False
        self.data_type = None
        self.enumerable_value = False
        self.value_attr = None
        self.value_handler = None
        self.attributes = None
        self._default = kwargs.pop('default', None)
        self.journaling = kwargs.pop('journaling', True)
        self.required = bool(kwargs.pop('required', False))
        self.nullable = bool(kwargs.pop('nullable', not self.required))
        self.l10n = kwargs.pop('l10n', None)
        self.override_null = None
        self.on_new = kwargs.pop('on_new', None)
        self.on_change = kwargs.pop('on_change', None)
        self.on_text_repr = kwargs.get('on_text_repr', None)
        self.on_log_repr = kwargs.get('on_log_repr', None)
        self.setter = kwargs.pop('setter', None)
        self.getter = kwargs.pop('getter', None)
        self.in_composite_results = True
        self._caption = kwargs.pop('caption', None)
        self._abbreviation = kwargs.pop('abbreviation', None)
        self._options = kwargs.pop('options', None)
        if self.on_new is not None and not callable(self.on_new):
            raise OrmParameterError("Attributes's `on_new` must be a callable function!")
        if self.on_change is not None and not callable(self.on_change):
            raise OrmParameterError("Attributes's `on_change` must be a callable function!")

    def on_before_write(self, instance):
        """ Calls for every Model's data attribute right before starting a write procedure
        of an instance. """
        pass

    def on_after_write(self, instance):
        """ Calls for every Model's data attribute after object instance been written to
        the database, as last action. """
        pass

    def on_before_delete(self, instance, hard=None):
        """ Calls for every Model's data attribute right before starting a deletion
        procedure of an instance. """
        pass

    def on_after_delete(self, instance, hard=None):
        """ Calls for every Model's data attribute after object instance deletion occures. """
        pass

    def _l10n_text(self, s):
        """ Tries to localize given text. """
        if isinstance(self.l10n, bool):
            return s if not self.l10n else gettext(s)
        return self.l10n(s) if callable(self.l10n) else s

    def initialize(self, instance, key):
        """ Initially sets the value of this attribute. By default - calls model's
        __setattrvalue__ method to store the value directly. Datatype, based on this
        abstract class, may override this method to make some actions to store
        value correctly depending on the its logic. """
        instance.__setattrvalue__(key, self.default)

    def get_default_value(self):
        """ Returns the default value for this attribute. While default value may be
        set_value using not only static value, but also a callable function - this method
        returns a result of that callable function if it is set_value. """
        if callable(self._default):
            return self._default()
        else:
            return self._default

    def get_default_definition(self):
        """ Returns the definition of default value. Difference between this method and
        'default' property that last one will be resolved to the end value if is a
        dynamic one (is a callable function). """
        return self._default

    @property
    def default(self):
        """ Returns resolved default value for this attribute. """
        return self.get_default_value()

    @default.setter
    def default(self, value):
        """ Sets the default value definition for this attribute. """
        self._default = value

    @property
    def caption(self):
        """ The textual full-sized caption of the attribute. Not used at DBMS side. """
        return self._get_f_value('_caption', None) or self._get_f_value('_abbreviation', None) or self.model_k

    @caption.setter
    def caption(self, value):
        """ Sets the caption of this attribute. """
        self._caption = value

    @property
    def abbreviation(self):
        """ The abbreviation of the attribute. Not used at DBMS side. """
        return self._get_f_value('_abbreviation', None) or self._get_f_value('_caption', None) or self.model_k

    @abbreviation.setter
    def abbreviation(self, value):
        """ Sets the abbreviation of the attribute. """
        self._abbreviation = value

    @property
    def options(self):
        """ Set of options for the attribute. """
        from ..request import OrmRequest, REQUEST_SELECT
        from ..models import OrmModel

        value = self._get_f_value('_options', None)
        if isinstance(value, dict) or value is None:
            return value
        if isinstance(value, OrmRequest):
            if value.request_type != REQUEST_SELECT:
                raise OrmParameterError(
                    "Cannot use given request for `options` because request must be a type of OrmSelectRequest -"
                    " it must return data from the database."
                )
            rows = value.go()
            value = dict()
            for row in rows:
                k_, v_ = row.__dictpair__
                value[k_] = v_
        elif isinstance(value, str) or isinstance(value, OrmModel):
            model = mapper.get_model_by_name(value) if isinstance(value, str) else value
            if model is None:
                raise OrmInvalidKey(
                    "There is no model with name `%s` used when declaring `options`!"
                    % value
                )
            value = model.directory()
        return value

    @options.setter
    def options(self, value):
        """ Sets the defition of the options of this attribute. """
        self._options = value

    def text_repr(self, value, instance=None):
        """ Used to return human readable textual representation of the raw data of this attribute. """
        # The textual representation can be set_value to the static value, without correlation
        # to the database loaded value.
        if isinstance(self.on_text_repr, str):
            return self._l10n_text(self.on_text_repr)

        # If there is a method defined to represent the value - use it at first point.
        if callable(self.on_text_repr):
            return self._l10n_text(self.on_text_repr(value))

        # If the options list is defined - using it to try to represent the value.
        if self._options is not None:
            options = self.options
            if not options or not isinstance(options, dict):
                return self._l10n_text(value)
            if value not in options:
                return self._l10n_text(value)
            return self._l10n_text(options[value])

        # At last - return value itself.
        return "" if value is None else self._l10n_text(value)

    def log_repr(self, src, dst, instance=None):
        """ Used to return human readable textual representation of the raw value of this
        column to store to the log (history) when writting log.
        @src:               An original value which was before change;
        @dst:               A target value which been set_value;
        @return:            (tuple|list) of two values: textual representation of `value_f`
                            and textual representation of `value_t`.
        """
        # The textual representation can be set_value to the static value, without correlation
        # to the database loaded value.
        if isinstance(self.on_log_repr, str):
            return self._l10n_text(self.on_log_repr), None

        # If there is a method defined to represent the value - use it at first point.
        if callable(self.on_log_repr):
            r = self.on_log_repr(src, dst)
            if not isinstance(r, (tuple, list)) or len(r) != 2:
                raise TypeError(
                    "Custom, assigned to the column definition `log_repr` function"
                    " must return a tuple or list of two values: a textual representation"
                    " of source(from) value and destination(to) one!"
                )
            return self._l10n_text(r[0]), self._l10n_text(r[1])

        # Otherwise - return text_repr as default
        return self.text_repr(src, instance), self.text_repr(dst, instance)

    def verify(self, value):
        """ Verifies the value which about to be set to this instance's attribute. This
        method can raise 'OrmValueVerificationError' with explataion text if the value
        cannot be set, and can transform or modify value to be set, returning the
        resulting value.
        This method not sets value, but only verifies it and returns what about to be
        actually set. """
        if isinstance(value, AttributeValueNull):
            value = None
        if not value and str(value) != '0' and str(value) != '0.0' and self.required:
            raise OrmValueVerificationError(gettext('Is required'))
        if value is None and not self.nullable:
            if self.override_null is not None:
                return self.override_null
            raise OrmValueVerificationError(gettext("Must be set"))
        return value


class ColumnAttributeAbstract(OrmComparatorMixin, DataAttributeAbstract):
    """ Abstract class for DBMS driven column types """

    def __init__(self, **kwargs):
        from .columns import ColumnIntegerAbstract
        super(ColumnAttributeAbstract, self).__init__(**kwargs)
        self.is_softdelete_key = kwargs.pop('is_softdelete_key', False)
        self.is_default_order = bool(kwargs.pop('is_default_order', False))
        self.is_version_key = bool(kwargs.pop('is_version_key', False))
        self.is_caption_key = bool(kwargs.pop('is_caption_key', False))
        self.is_value_key = bool(kwargs.pop('is_value_key', False))
        self.extender_key = bool(kwargs.pop('extender_key', False))
        self.primary_key = bool(kwargs.pop('primary_key', False))
        self.auto_increment = bool(kwargs.pop('auto_increment', False))
        self.unique = kwargs.pop('unique', None)
        self.key = kwargs.pop('key', None)
        self.comment = kwargs.pop('comment', None)
        self.findable = bool(kwargs.pop('findable', False))
        self.strict_find = bool(kwargs.pop('strict_find', False))
        self.manual = bool(kwargs.pop('manual', False))
        self.references = kwargs.pop('references', None)  # reference using auto-FK on column of another OrmModel
        self.protected = bool(kwargs.pop('protected', False))
        if self.auto_increment and not isinstance(self, ColumnIntegerAbstract):
            raise OrmAttributeTypeMistake("AUTO_INCREMENT column must be the integer type!")
        if self.auto_increment:
            self.primary_key = True
            self.nullable = False
            self._default = None
        self._on_get = kwargs.pop('on_get', None)
        self._on_set = kwargs.pop('on_set', None)
        self._cast_write = kwargs.pop('cast_write', None)
        self._cast_read = kwargs.pop('cast_read', None)
        self._cast_order_by = kwargs.pop('cast_order_by', self._cast_read)
        self._cast_group_by = kwargs.pop('cast_group_by', self._cast_read)
        self._cast_filter = kwargs.pop('cast_filter', self._cast_read)
        for n in kwargs:
            if hasattr(self, n):
                continue
            v = kwargs.get(n)
            setattr(self, n, v)

    def __hash__(self):
        return super(OrmComparatorMixin, self).__hash__()

    def as_(self, name):
        """ Returns this attribute formatted to be named column as `name`. """
        return AttributeAliased(self, name)

    @staticmethod
    def _cmp_c_definition(val1, val2):
        if val1 is None and val2 is not None:
            return False
        if val1 is not None and val2 is None:
            return False
        return val1 == val2

    def alter_sql(self, session):
        pass

    def declaration_sql(self, session):
        pass

    def is_eq_with_tbl_schema(self, struct, session):
        pass

    def represent_for_write(self, value, session):
        return value

    def make_write(self, context):
        pn = context.get_valueparam_by_column(self)
        if not pn:
            return None
        if self._cast_write is not None:
            return self._cast_write(context, pn)
        if self.auto_increment and not context.params[pn]:
            return None
        return "%(" + pn + ")s"

    def make_read(self, context):
        if not self._cast_read:
            return self
        return self._cast_read(context)

    def make_filter(self, context):
        if not self._cast_filter:
            return self
        return self._cast_filter(context)

    def make_order_by(self, context):
        if not self._cast_order_by:
            return self.resolve_opts_to_case()
        return self._cast_order_by(context)

    def make_group_by(self, context):
        if not self._cast_group_by:
            return self
        return self._cast_group_by(context)

    def resolve_opts_to_case(self):
        from ..funcs import case_else_
        options = self.options
        if not options:
            return self
        fargs = list()
        for k in options:
            v = options[k]
            fargs.append((k, v))
        return case_else_(self, "", *fargs)

    @staticmethod
    def _find_prepare(value):
        return value

    def find_prepare(self, value):
        if isinstance(value, (list, tuple)):
            r = list()
            for val in value:
                v = self._find_prepare(val)
                if not v:
                    continue
                r.append(str(v))
            return r
        else:
            return self._find_prepare(value)


class VirtualAttributeAbstract(DataAttributeAbstract):
    """ Abstract class for virtual attributes, not directly reflected at DBMS """
    def write(self, instance, key):
        pass


class KeyAttributeAbstract(AttributeAbstract):
    def __init__(self, *args):
        if not args:
            raise OrmParameterError("No _attrs are specified when declaring KEY `%s`!" % self.model_k)
        super(KeyAttributeAbstract, self).__init__()
        self._validated = False
        self._args = args
        self.columns = list()

    def validate(self):
        if self._validated:
            return
        for arg in self._args:
            if not isinstance(arg, (str, ColumnAttributeAbstract)):
                raise OrmParameterError(
                    "When declaring KEYs: _attrs must be set using textual names (str) or attributes themselfs."
                )
            if isinstance(arg, str):
                if not hasattr(self.model, arg):
                    raise OrmInvalidKey(
                        "OrmModel `%s` has no attribute `%s` set for the key `%s`!" %
                        (self.model.__name__, arg, self.model_k)
                    )
                c = getattr(self.model, arg)
                if not isinstance(c, ColumnAttributeAbstract):
                    raise OrmAttributeTypeMistake(
                        "Attribute `%s` in the OrmModel `%s`, declaring in the key `%s`, is not a type of"
                        " Column! Only database table refrected attributes may be setattr for keys!" %
                        (arg, self.model.__name__, self.model_k)
                    )
                self.columns.append(arg)
            else:
                if arg.model != self.model:
                    raise OrmAttributeTypeMistake(
                        "Attribute `%s` is not belongs to the OrmModel `%s`, declaring in the key `%s`!"
                        " Only local OrmModel's attributes may be set for OrmModel's keys!" %
                        (arg, self.model.__name__, self.model_k)
                    )
                self.columns.append(arg.model_k)
        self._validated = True

    def get_column_names(self, quoted=False, session=None):
        """ Returns the list() of names of all keys of this unique keys """
        names = []
        for k in self.columns:
            c = getattr(self.model, k)
            n = c.tbl_k
            if quoted:
                n = session.quote_name(n)
            names.append(n)
        return names


class AttributeAliased(object):
    """ Aliases the attribute with another name. Useful with <>.as_() method, for example """
    def __init__(self, c, alias):
        self.c = c
        self.alias = alias

    def __repr__(self):
        return "AlisedAttr(%r AS '%s')" % (self.c, self.alias)


def _attribute(c, **kwargs):
    """ A virtual attribute declaration which receives an existing actual attribute
    definition as argument and modifiers as named arguments. Used to set_value up
    an existing attribute somewhere else, with posibility to redefine some
    attribute parameters.
    @c:             An existing attribute, already been declared before;
    @kwargs:        Optional modifiers which may change some parameters of the
                    newly declared attribute, which differs from the original one.
    """
    mc = copy.copy(c)
    for k in kwargs:
        setattr(mc, k, kwargs[k])
    return mc


Attribute = _attribute


def _caption(*args, **kwargs):
    from .derived import CaptionString
    if args and isinstance(args[0], AttributeAbstract):
        return Attribute(args[0], is_caption_key=True, **kwargs)
    return CaptionString(*args, **kwargs)


Caption = _caption




# -*- coding: utf8 -*-
from pc23 import metaclass
from .meta import OrmViewMeta
from ..funcs import and_, where_, having_, RequestOrderBy, RequestGroupBy


@metaclass(OrmViewMeta)
class OrmView(object):
    @classmethod
    def __attrsdir__(cls):
        keys = cls.__attrs__
        attrs = list()
        for k in keys:
            c = getattr(cls, k)
            attrs.append((c.declaration_order, k))
        return [y[1] for y in sorted(attrs, key=lambda x: x[0])]

    @classmethod
    def _get_where(cls):
        def_grnd = getattr(cls, '__where__', None)
        def_attr = getattr(cls, 'where', None)
        if not def_grnd and not def_attr:
            return None
        r = list()
        if def_grnd:
            r.append(def_grnd)
        if def_attr:
            r.append(def_attr)
        return r if len(r) == 1 else and_(*r)

    @classmethod
    def _get_having(cls):
        def_grnd = getattr(cls, '__having__', None)
        def_attr = getattr(cls, 'having', None)
        if not def_grnd and not def_attr:
            return None
        r = list()
        if def_grnd:
            r.append(def_grnd)
        if def_attr:
            r.append(def_attr)
        return r if len(r) == 1 else and_(*r)

    @classmethod
    def _get_order(cls):
        def_grnd = getattr(cls, '__order__', None)
        def_attr = getattr(cls, 'order', None)
        if not def_grnd and not def_attr:
            return None
        r = def_attr if def_attr else def_grnd
        if isinstance(r, RequestOrderBy):
            return [r, ]
        if not isinstance(r, (list, tuple)):
            return [RequestOrderBy(r), ]
        rr = list()
        for r_ in r:
            rr.append(RequestOrderBy(*r_) if isinstance(r_, (list, tuple)) else RequestOrderBy(r_))
        return rr

    @classmethod
    def _get_group(cls):
        def_grnd = getattr(cls, '__group__', None)
        def_attr = getattr(cls, 'group', None)
        if not def_grnd and not def_attr:
            return None
        r = def_attr if def_attr else def_grnd
        if isinstance(r, RequestGroupBy):
            return [r, ]
        if not isinstance(r, (list, tuple)):
            return [RequestGroupBy(r), ]
        rr = list()
        for r_ in r:
            rr.append(RequestGroupBy(*r_) if isinstance(r_, (list, tuple)) else RequestGroupBy(r_))
        return rr

    @classmethod
    def _request_skeleton(cls, *args):
        from ..funcs import left_join_, t_
        from ..attributes.base import AttributeAbstract
        model = getattr(cls, '__model__')
        attrs_unsorted = getattr(cls, '__attrs__')
        joins_unsorted = getattr(cls, '__joins__')
        attrs = list()
        joins = list()
        for k in attrs_unsorted:
            c = getattr(cls, k)
            attrs.append((c.declaration_order, k))
        for k in joins_unsorted:
            c = getattr(cls, k)
            joins.append((c.declaration_order, k))
        attrs = [y[1] for y in sorted(attrs, key=lambda x: x[0])]
        joins = [y[1] for y in sorted(joins, key=lambda x: x[0])]

        _args = list()
        _args.append(model)
        for k in attrs:
            c = getattr(cls, k)
            if c.func is not None:
                _args.append(c.func.as_(k))
            elif c.owner == model.__name__ or c.owner is None:
                _args.append(getattr(model, c.owner_k).as_(k))
            elif c.owner in joins:
                _args.append(getattr(t_(c.owner), c.owner_k).as_(k))
            else:
                continue

        for k in joins:
            c = getattr(cls, k)
            jargs = list()
            for ja in c.args:
                if isinstance(ja.item, AttributeAbstract) and ja.item.model == c.ref_model:
                    ja.item = getattr(t_(k), ja.item.model_k)
                if isinstance(ja.value, AttributeAbstract) and ja.value.model == c.ref_model:
                    ja.value = getattr(t_(k), ja.value.model_k)
                jargs.append(ja)
            _args.append(left_join_(c.ref_model.as_(k), *jargs))

        where = cls._get_where()
        if where:
            _args.append(where_(*where))
        having = cls._get_having()
        if having:
            _args.append(having_(*having))
        order = cls._get_order()
        if order:
            for o in order:
                _args.append(o)
        group = cls._get_group()
        if group:
            for o in group:
                _args.append(o)

        for arg in args:
            _args.append(arg)
        return _args

    @classmethod
    def get(cls, *args, **kwargs):
        return cls.request(*args, **kwargs).go()

    @classmethod
    def find(cls, what, **kwargs):
        from ..request import OrmSearchRequest
        _args = cls._request_skeleton()
        _args.insert(1, what)
        return OrmSearchRequest(*_args, **kwargs)

    @classmethod
    def request(cls, *args, **kwargs):
        from ..request import OrmSelectRequest
        return OrmSelectRequest(*cls._request_skeleton(*args), **kwargs)

